"""Define basic schema for band theoretic properties."""

from __future__ import annotations

import numbers
from typing import TYPE_CHECKING

import numpy as np
from pydantic import BaseModel, Field, computed_field, field_validator
from pymatgen.core import Lattice
from pymatgen.electronic_structure.bandstructure import (
    BandStructure as PmgBandStructure,
)
from pymatgen.electronic_structure.bandstructure import Kpoint
from pymatgen.electronic_structure.core import Orbital, Spin
from pymatgen.electronic_structure.dos import CompleteDos, Dos

from emmet.core.math import Matrix3D, Vector3D
from emmet.core.types.pymatgen_types.structure_adapter import StructureType

if TYPE_CHECKING:

    from pymatgen.core.sites import PeriodicSite
    from typing_extensions import Self

BAND_GAP_TOL = 1e-4


class BandTheoryBase(BaseModel):

    identifier: str | None = Field(None, description="The identifier of this object.")
    structure: StructureType | None = Field(
        None, description="The structure associated with this calculation."
    )


class BandStructure(
    BandTheoryBase, populate_by_name=True, validate_by_alias=True, validate_by_name=True
):
    """Define the schema of band structures.

    This class is generic enough to accommodate both
    electronic and phonon band structures.
    """

    qpoints: list[Vector3D] = Field(
        description="The wave vectors (q-points) at which the band structure was sampled, in direct coordinates.",
    )

    reciprocal_lattice: Matrix3D = Field(
        description="The reciprocal lattice.", validation_alias="lattice_rec"
    )

    labels_dict: dict[str, Vector3D] = Field(
        {}, description="The high-symmetry labels of specific q-points."
    )

    @field_validator("reciprocal_lattice", mode="before")
    def reciprocal_lattice_deserializer(cls, reciprocal_lattice):
        if isinstance(reciprocal_lattice, dict):
            return reciprocal_lattice.get("matrix")

        return reciprocal_lattice


class ElectronicBS(BandStructure):
    """Define an electronic band structure schema."""

    efermi: float = Field(
        description="The Fermi level (highest occupied energy level.)"
    )

    band_gap: float | None = Field(None, description="The value of the band gap.")

    spin_up_bands: list[list[float]] | None = Field(
        None,
        description="The eigen-energies of the spin-up electrons. The first index represents the band, the second the k-point index.",
    )

    spin_down_bands: list[list[float]] | None = Field(
        None,
        description="The eigen-energies of the spin-down electrons. The first index represents the band, the second the k-point index.",
    )

    is_direct: bool | None = Field(
        None, description="If the bandgap is non-zero, whether the band gap is direct."
    )

    spin_up_projections: list[list[list[list[float]]]] | None = Field(
        None,
        description=(
            "The atom-projected eigen-energies of the spin-up electrons. "
            "The first index is the band index. The second index is the k-point index. "
            "The third index is the site index. The final index is the orbital character, "
            "ordered as s, py, pz, px, dxy, dyz, dz2, dxz, dx2-y2,..."
        ),
    )

    spin_down_projections: list[list[list[list[float]]]] | None = Field(
        None,
        description=(
            "The atom-projected eigen-energies of the spin-down electrons. "
            "The first index is the band index. The second index is the k-point index. "
            "The third index is the site index. The final index is the orbital character, "
            "ordered as s, py, pz, px, dxy, dyz, dz2, dxz, dx2-y2,..."
        ),
    )

    @computed_field
    def is_metal(self) -> bool:
        """Whether the band gap is almost zero."""
        return (
            self.band_gap < BAND_GAP_TOL
            if isinstance(self.band_gap, numbers.Number)
            else False
        )

    @classmethod
    def from_pmg(cls, ebs: PmgBandStructure) -> Self:
        """Construct from a pymatgen band structure object."""
        band_gap_meta = ebs.get_band_gap()
        config = {
            "qpoints": [qpt.frac_coords for qpt in ebs.kpoints],
            "lattice_rec": ebs.lattice_rec.matrix,
            "efermi": ebs.efermi,
            "labels_dict": {
                label: qpt.frac_coords for label, qpt in ebs.labels_dict.items()
            },
            "structure": ebs.structure,
            "is_metal": ebs.is_metal(),
            "is_direct": band_gap_meta["direct"],
            "band_gap": band_gap_meta["energy"],
        }

        for spin in Spin:
            config[f"spin_{spin.name}_bands"] = ebs.bands.get(spin)
            if ebs.projections is not None:
                config[f"spin_{spin.name}_projections"] = ebs.projections.get(spin)
        return cls(**config)

    def to_pmg(
        self,
    ) -> PmgBandStructure:
        """Construct the pymatgen object from the current instance."""
        rlatt = Lattice(self.reciprocal_lattice)

        bands = {}
        projections = {}
        for spin in Spin:
            if v := getattr(self, f"spin_{spin.name}_bands", None):
                bands[spin] = np.array(v)
            if v := getattr(self, f"spin_{spin.name}_projections", None):
                projections[spin] = np.array(v)

        return PmgBandStructure(
            [Kpoint(q, lattice=rlatt).frac_coords for q in self.qpoints],  # type: ignore[misc]
            bands,
            rlatt,
            self.efermi,
            labels_dict={
                k: Kpoint(v, lattice=rlatt).frac_coords  # type: ignore[misc]
                for k, v in (self.labels_dict or {}).items()
            },
            coords_are_cartesian=False,
            structure=self.structure,
            projections=projections,
        )


class BaseElectronicDos(BandTheoryBase):
    """Basic structure for spin-resolved density of states (DOS).

    Note that this class is intended for storing:
        1. Total DOS.
        2. A single-orbital-resolved component of a site- and orbital-
            projected DOS.
    """

    spin_up_densities: list[float] | None = Field(None, description="The spin-up DOS.")

    spin_down_densities: list[float] | None = Field(
        None, description="The spin-down DOS."
    )

    energies: list[float] | None = Field(
        None, description="The energies at which the DOS was calculated."
    )

    efermi: float | None = Field(None, description="The Fermi energy.")

    orbital: str | None = Field(
        None, description="The orbital character of this DOS, if applicable."
    )

    def to_pmg(self) -> Dos:
        """Convert to a pymatgen DOS object."""
        densities = {}
        for spin in Spin:
            if sr_density := getattr(self, f"spin_{spin.name}_densities", None):
                densities[spin] = sr_density
        return Dos(self.efermi, self.energies, densities)  # type: ignore[arg-type]


class ElectronicDos(BaseElectronicDos):
    """Electronic density of states (DOS) with possible projections."""

    projected_densities: list[list[BaseElectronicDos] | None] | None = Field(
        None, description="The orbital- and site-projected DOS."
    )

    @property
    def _available_spins(self) -> list[Spin]:
        """Retrive spin indices of non-null spin data."""
        return [
            spin
            for spin in Spin
            if getattr(self, f"spin_{spin.name}_densities", None) is not None
        ]

    @classmethod
    def from_pmg(cls, dos: Dos | CompleteDos, **kwargs) -> Self:
        """Create an electronic DOS from a pymatgen object."""

        densities: dict[str, list[float]] = {
            f"spin_{spin.name}_densities": list(sr_dos)
            for spin, sr_dos in dos.densities.items()
        }

        if isinstance(dos, CompleteDos):

            pdos: list[list[BaseElectronicDos] | None] | None = None
            if (vrun_pdos := dos.pdos) and dos.structure:
                pdos_dct: dict[int, list[BaseElectronicDos] | None] = {}
                for site, orb_spin_dos in vrun_pdos.items():
                    site_idx: int = [
                        idx
                        for idx, ref_site in enumerate(dos.structure)
                        if ref_site == site
                    ][0]
                    pdos_dct[site_idx] = [
                        BaseElectronicDos(
                            **{  # type: ignore[arg-type]
                                f"spin_{spin.name}_densities": sr_dos
                                for spin, sr_dos in spin_dos.items()
                            },
                            orbital=orbital.name,
                        )
                        for orbital, spin_dos in orb_spin_dos.items()
                    ]
                pdos = [pdos_dct[idx] for idx in sorted(pdos_dct)]

        return cls(
            **densities,  # type: ignore[arg-type]
            efermi=dos.efermi,
            energies=dos.energies.tolist(),
            structure=getattr(dos, "structure", None),
            projected_densities=pdos,
            **kwargs,
        )

    def to_pmg(self) -> Dos | CompleteDos:
        """Serialize to pymatgen."""
        if self.efermi is None:
            raise ValueError(
                "Fermi level unspecified, cannot create a pymatgen (Complete)Dos object."
            )

        densities: dict[Spin, np.ndarray] = {}
        for spin in self._available_spins:
            if (
                _dens := getattr(self, f"spin_{spin.name}_densities", None)
            ) is not None:
                densities[spin] = np.array(_dens)

        dos = Dos(self.efermi, np.array(self.energies), densities)
        if self.structure and self.projected_densities:

            pdos: dict[PeriodicSite, dict[Orbital, dict[Spin, np.ndarray]]] = {}
            for isite, site in enumerate(self.structure):
                pdos[site] = {}
                if not self.projected_densities[isite]:
                    continue
                for site_dos in self.projected_densities[
                    isite
                ]:  # type:ignore[union-attr]
                    if not site_dos.orbital:
                        continue
                    orbital = Orbital[site_dos.orbital]
                    pdos[site][orbital] = {}
                    for spin in self._available_spins:
                        if (
                            _pdos := getattr(
                                site_dos, f"spin_{spin.name}_densities", None
                            )
                        ) is not None:
                            pdos[site][orbital][spin] = np.array(_pdos)

            return CompleteDos(self.structure, dos, pdos)

        return dos



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.core import Composition

CompositionTypeVar = TypeVar("CompositionTypeVar", Composition, dict[str, float])


def composition_deserializer(comp):
    if isinstance(comp, list):
        comp = {k: v for k, v in comp}

    if isinstance(comp, dict):
        return Composition.from_dict(comp)

    return comp


CompositionType = Annotated[
    CompositionTypeVar,
    BeforeValidator(lambda x: Composition.from_dict(x) if isinstance(x, dict) else x),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=dict[str, float]),
]



"""Define schemas for trajectories."""

from __future__ import annotations

import logging
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from monty.dev import requires
from monty.serialization import dumpfn
from pydantic import BaseModel, Field, model_validator
from pymatgen.core import Element, Molecule, Structure
from pymatgen.core.trajectory import Trajectory as PmgTrajectory

from emmet.core.math import Matrix3D, Vector3D
from emmet.core.vasp.calc_types import RunType, TaskType, calc_type, run_type, task_type
from emmet.core.vasp.calculation import ElectronicStep
from emmet.core.types.enums import ValueEnum

logger = logging.getLogger(__name__)

try:
    import pyarrow as pa
    import pyarrow.parquet as pa_pq
    from pyarrow import Table as ArrowTable

except ImportError:
    pa = None
    pa_pq = None
    ArrowTable = None

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any

    from typing_extensions import Self

    from emmet.core.vasp.calc_types import CalcType
    from emmet.core.vasp.calculation import Calculation


class TrajFormat(ValueEnum):
    """Define known trajectory formats."""

    PMG = "json"
    ASE = "traj"
    PARQUET = "parquet"


class _MDMixin(BaseModel):
    """Mix-in for molecular dynamics runs."""

    magmoms: list[list[float]] | None = Field(
        None, description="The magnetic moments at each ionic step."
    )

    temperature: list[float] | None = Field(
        None, description="The temperature at each ionic step."
    )

    velocities: list[list[Vector3D]] | None = Field(
        None, description="The velocities of each atom."
    )

    ionic_step_properties: set[str] = Field(
        {"magmoms", "temperature", "velocities"},
        exclude=True,
        description="The properties included at each ionic step.",
    )

    @model_validator(mode="after")
    def add_props(self) -> Self:
        self.ionic_step_properties = self.ionic_step_properties.union(
            {"magmoms", "temperature", "velocities"}
        )
        return self


class AtomRelaxTrajectory(BaseModel):
    """Atomistic only, low-memory schema for relaxation trajectories that can interface with parquet, pymatgen, and ASE."""

    elements: list[int] = Field(
        description="The proton number Z of the elements in the sites"
    )

    cart_coords: list[list[Vector3D] | None] = Field(
        description="The Cartesian coordinates (Å) of the sites at each ionic step"
    )
    num_ionic_steps: int = Field(description="The number of ionic steps.")

    lattice: list[Matrix3D] | None = Field(
        None,
        description=(
            "If a list containing only one 3x3 matrix, it is assumed that "
            "the lattice was held constant through the simulation. "
            "If a list of (multiple) 3x3 matrices, this should be the lattice "
            "at each ionic step in the calculation. "
            "If None, a non-periodic system is assumed."
        ),
    )

    energy: list[float] | None = Field(
        None, description="The total energy at each ionic step."
    )

    forces: list[list[Vector3D]] | None = Field(
        None, description="The interatomic forces."
    )
    stress: list[Matrix3D] | None = Field(None, description="The 3x3 stress tensor.")

    ionic_step_properties: set[str] = Field(
        {
            "energy",
            "forces",
            "stress",
        },
        exclude=True,
        description="The properties included at each ionic step.",
    )

    @model_validator(mode="before")
    def reset_num_ionic_steps(cls, config: Any) -> Any:
        config["num_ionic_steps"] = len(config["cart_coords"])
        config["elements"] = [
            Element(ele).Z if isinstance(ele, str) else int(ele)
            for ele in config["elements"]
        ]
        return config

    def __hash__(self) -> int:
        """Used to verify roundtrip conversion of Trajectory."""
        return hash(self.model_dump_json())

    def __len__(self) -> int:
        """Get number of ionic steps."""
        return self.num_ionic_steps

    def __getitem__(self, index: int | slice) -> AtomRelaxTrajectory:
        if isinstance(index, int):
            if index < 0:
                index = len(self) + index
        _slice: slice = slice(index, index + 1) if isinstance(index, int) else index

        config = {}
        all_ionic_step_props = self.ionic_step_properties.union(
            {
                "lattice",
                "cart_coords",
            }
        )
        for k in self.__class__.model_fields:
            v = getattr(self, k)
            if v and k in all_ionic_step_props:
                config[k] = v[_slice]
            elif v:
                config[k] = v
        return type(self)(**config)

    @staticmethod
    def reorder_sites(
        structure: Structure | Molecule, ref_z: list[int]
    ) -> tuple[Structure | Molecule, list[int]]:
        """
        Ensure that the sites in a structure match the order in a set of reference Z values.

        Quick returns if the structure already matches the reference order.

        Parameters
        -----------
        structure : pymatgen .Structure or .Molecule
            Structure or Molecule to order sites in
        ref_z : list of int
            List of proton numbers / Z values in the reference structure.

        Returns
        -----------
        pymatgen .Structure or .Molecule : structure matching the reference order of sites
        list of int : indices in the original structure which are reordered
        """

        is_structure = isinstance(structure, Structure)

        if [site.species.elements[0].Z for site in structure] == ref_z:
            return structure, list(range(structure.num_sites))

        running_sites = list(range(len(structure)))
        ordered_sites = []
        reordered_index = []
        for z in ref_z:
            for idx in running_sites:
                if structure[idx].species.elements[0].Z == z:
                    ordered_sites.append(structure[idx])
                    reordered_index.append(idx)
                    running_sites.remove(idx)
                    break

        if len(running_sites) > 0:
            raise ValueError(
                f"Cannot order {'structure' if is_structure else 'molecule'} according to reference list."
            )

        return (Structure if is_structure else Molecule).from_sites(
            ordered_sites
        ), reordered_index

    @classmethod
    def _from_dict(
        cls,
        props: dict[str, Any],
        constant_lattice: bool = False,
        lattice_match_tol: float | None = None,
        **kwargs,
    ) -> Self:
        """
        Common class constructor to create a Trajectory from a dict.

        Parameters
        -----------
        props : dict of str to list
            Dictionary with keys in cls.model_fields and values which are lists
        constant_lattice : bool (default = False)
            Whether the lattice is constant thoughout the trajectory
        lattice_match_tol : float or None (default = 1e-6)
            If a float and constant_lattice is False, this defines the tolerance
            for determining if the lattice has changed during the course of the
            trajectory. If None, no check is performed.
        **kwargs
            Other kwargs to pass to Trajectory

        Returns
        -----------
        Trajectory
        """
        structures = props.pop("structure")
        is_structure = isinstance(structures[0], Structure)
        if is_structure and any(not structure.is_ordered for structure in structures):
            raise ValueError(
                "At this time, the Trajectory model in emmet does not support disordered structures."
            )

        num_ionic_steps = len(structures)

        props["elements"] = [site.species.elements[0].Z for site in structures[0]]
        if (
            is_structure
            and constant_lattice
            or (
                lattice_match_tol
                and all(
                    np.all(
                        np.abs(
                            structures[i].lattice.matrix - structures[j].lattice.matrix
                        )
                        < lattice_match_tol
                    )
                    for i in range(num_ionic_steps)
                    for j in range(i + 1, num_ionic_steps)
                )
            )
        ):
            props["lattice"] = [structures[0].lattice.matrix]
        elif is_structure:
            props["lattice"] = [structure.lattice.matrix for structure in structures]

        props["cart_coords"] = [
            cls.reorder_sites(structure, props["elements"])[0].cart_coords
            for structure in structures
        ]

        if len(esteps := props.get("electronic_steps", [])) > 0:
            props["num_electronic_steps"] = [len(estep) for estep in esteps]

        return cls(**props, **kwargs)

    @classmethod
    def from_pmg(cls, traj: PmgTrajectory, **kwargs) -> Self:
        """
        Create a trajectory from a pymatgen .Trajectory.

        Parameters
        -----------
        traj : pymatgen.core.trajectory.Trajectory
        constant_lattice : bool (default = False)
            Whether the lattice is constant thoughout the trajectory
        lattice_match_tol : float or None (default = 1e-6)
            If a float and constant_lattice is False, this defines the tolerance
            for determining if the lattice has changed during the course of the
            trajectory. If None, no check is performed.

        Returns
        -----------
        Trajectory
        """
        props: dict[str, list] = {
            "structure": [structure for structure in traj],
        }
        if traj.frame_properties:
            for k in cls.model_fields:
                vals = [fp.get(k) for fp in traj.frame_properties]
                if all(v is not None for v in vals):
                    props[k] = vals

        return cls._from_dict(props, **kwargs)

    @requires(
        pa is not None, message="pyarrow must be installed to de-/serialize to parquet"
    )
    def to_arrow(
        self,
        file_name: str | Path | None = None,
        **write_file_kwargs,
    ) -> ArrowTable:
        """
        Create a PyArrow Table from a Trajectory.

        Parameters
        -----------
        file_name : str, .Path, or None (default)
            If not None, a file to write the parquet-format output to.
            Accepts any compression extension used by pyarrow.write_table
        **write_file_kwargs
            If file_name is not None, any kwargs to pass to
            pyarrow.parquet.write_file

        Returns
        -----------
        pyarrow.Table
        """
        pa_config = {
            k: pa.array([v])
            for k, v in self.model_dump(mode="json").items()
            if hasattr(v, "__len__")
        }

        pa_table = pa.table(pa_config)
        if file_name:
            pa_pq.write_table(pa_table, file_name, **write_file_kwargs)

        return pa_table

    @classmethod
    def from_arrow(cls, pa_table: ArrowTable, identifier: str | None = None) -> Self:
        """
        Create a trajectory from an arrow Table.

        Parameters
        -----------
        file_name : str or .Path
            The parquet file to read from
        identifier : str or None (default)
            If not None, the identifier of the trajectory to return
            from a dataset of multiple trajectories.

        Returns
        -----------
        Trajectory
        """
        if identifier:
            pa_table = pa_table.filter(
                pa.compute.field("identifier") == identifier,
                null_selection_behavior="drop",
            )

        config = {
            k: pa_table[k].to_pylist()[0]
            for k in pa_table.column_names
            if k in cls.model_fields
        }
        return cls(**config)

    @classmethod
    def from_parquet(cls, file_name: str | Path, identifier: str | None = None) -> Self:
        """Create a Trajectory from a parquet file.

        Parameters
        -----------
        file_name : str | Path
            Path to the parquet file. Can be a remote path, e.g. AWS.
        identifier : str | None = None
            The string identifier for the task.
        """
        kwargs = {}
        if identifier:
            kwargs["filters"] = [("identifier", "=", identifier)]
        return cls.from_arrow(pa_pq.read_table(file_name, **kwargs))

    def to_pmg(
        self,
        frame_props: Iterable[str] | None = None,
        indices: int | slice | Iterable[int] | None = None,
    ) -> PmgTrajectory:
        """
        Create a pymatgen Trajectory.

        Parameters
        -----------
        frame_props : Sequence of str or None (default)
            If not None, a list of model fields to populate the frame properties
            of the pymatgen Trajectory with.
            If None, defaults to all available fields.

        Returns
        -----------
        pymatgen.core.trajectory.Trajectory
        """
        frame_prop_keys: set[str] = set(frame_props or self.ionic_step_properties)
        if "magmoms" in frame_prop_keys:
            frame_prop_keys.discard("magmoms")

        if indices:
            if isinstance(indices, slice):
                indices = set(range(indices.start, indices.stop, indices.step or 1))
            elif isinstance(indices, int):
                indices = {
                    indices,
                }
        else:
            indices = set(range(len(self)))
        indices = {i for i in indices if self.cart_coords[i]}

        species = [Element.from_Z(z) for z in self.elements]
        structures = []
        frame_properties = []

        for i in indices:
            site_properties = {}
            if magmoms := getattr(self, "magmoms", None):
                site_properties["magmoms"] = magmoms[i]

            if self.lattice:
                structure: Structure | Molecule = Structure(
                    lattice=(
                        self.lattice[0]
                        if len(self.lattice) == 1
                        else self.lattice[i]  # type: ignore[index]
                    ),
                    species=species,
                    coords=self.cart_coords[i],  # type: ignore[arg-type]
                    coords_are_cartesian=True,
                    site_properties=site_properties,
                )
            else:
                structure = Molecule(
                    species=species,
                    coords=self.cart_coords[i],  # type: ignore[arg-type]
                    site_properties=site_properties,
                )
            structures.append(structure)

            props = {}
            if len(frame_prop_keys) > 0:
                for k in frame_prop_keys:
                    if _prop := getattr(self, k, []):
                        prop = _prop[i]
                        for cmth in ("model_dump", "tolist"):
                            if hasattr(prop, cmth):
                                prop = getattr(prop, cmth)()
                        props[k] = prop

                frame_properties.append(props)

        if self.lattice:
            return PmgTrajectory.from_structures(
                structures,  # type: ignore[arg-type]
                constant_lattice=len(self.lattice) == 1,
                frame_properties=frame_properties,
            )
        return PmgTrajectory.from_molecules(
            structures,  # type: ignore[arg-type]
            frame_properties=frame_properties,
        )

    def to(
        self,
        file_name: str | Path | None = None,
        fmt: TrajFormat | str | None = None,
        **kwargs,
    ):
        """
        Generic interface to multiple trajectory formats.

        file_name : str, .Path, or None (default)
            If not None, the name of the file to write to.
            If fmt is None, the file format will be inferred from this.
        fmt : TrajFormat, str, or None (default)
            The format of the output trajectory. If file_name is None,
            and fmt is None, defaults to a parquet-compatible
            PyArrow table.
        **kwargs
            Any kwargs supported by Trajectory.to_* methods.
        """
        if file_name and not fmt:
            for _fmt in TrajFormat:  # type: ignore[attr-defined]
                if _fmt.value in str(file_name).lower():
                    fmt = _fmt
                    break
        elif not file_name and not fmt:
            fmt = TrajFormat.PARQUET

        if isinstance(fmt, str) and fmt.upper() in TrajFormat.__members__:
            fmt = TrajFormat[fmt.upper()]  # type: ignore[misc]
        else:
            fmt = TrajFormat(fmt)

        if fmt in (TrajFormat.PMG, TrajFormat.ASE):
            traj = self.to_pmg(**kwargs)

            if fmt == TrajFormat.PMG:
                if file_name:
                    dumpfn(traj, file_name)
            elif fmt == TrajFormat.ASE:
                if hasattr(PmgTrajectory, "to_ase"):
                    traj = traj.to_ase(ase_traj_file=file_name)  # type: ignore[attr-defined]
                else:
                    raise ImportError(
                        "Ensure you have pymatgen>=2025.1.23 to use ASE trajectory interfaces."
                    )

        elif fmt == TrajFormat.PARQUET:
            traj = self.to_arrow(file_name=file_name)

        return traj


class AtomTrajectory(AtomRelaxTrajectory, _MDMixin):
    """Atomistic trajectory with extra keys for molecular dynamics runs."""


class RelaxTrajectory(AtomRelaxTrajectory):
    """Low memory schema for relaxation trajectories that can interface with parquet, pymatgen, and ASE.

    This schema is intended primarily for VASP relaxation trajectories, but could be adapted to
    generic electronic structure trajectories.

    Note that for VASP calculations, energies are in eV, forces in eV/Å, and stress tensors in kilobar.
    """

    electronic_steps: list[list[ElectronicStep]] | None = Field(
        None, description="The electronic steps within a given ionic step."
    )
    num_electronic_steps: list[int] | None = Field(
        None, description="The number of electronic steps within each ionic step."
    )

    e_wo_entrp: list[float] | None = Field(
        None,
        description="The total energy in eV without electronic pseudoentropy from smearing of the Fermi surface.",
    )
    e_fr_energy: list[float] | None = Field(
        None, description="The total electronic free energy in eV."
    )

    identifier: str | None = Field(
        None, description="Identifier of this trajectory, e.g., task ID."
    )

    task_type: TaskType | None = Field(
        None,
        description="The TaskType of the calculation used to generate this trajectory.",
    )

    run_type: RunType | None = Field(
        None,
        description="The RunType of the calculation used to generate this trajectory.",
    )

    ionic_step_properties: set[str] = Field(
        {
            "energy",
            "forces",
            "stress",
            "e_wo_entrp",
            "e_fr_energy",
            "electronic_steps",
        },
        description="The properties included at each ionic step.",
        exclude=True,
    )

    @property
    def has_full_output(self) -> bool:
        """Return true if a trajectory has all structures and SCF convergence info available."""
        return all(coord_set is not None for coord_set in self.cart_coords) and all(
            conv_list is not None for conv_list in self.convergence_data.values()
        )

    @property
    def convergence_data(self) -> dict[str, list[float] | None]:
        """Get convergence of energy and interatomic forces.

        If possible, energy convergence is taken at every electronic step.
        If not, it is taken at each ionic step (the final electronic step).

        Forces are taken at each ionic step.
        """
        conv_data: dict[str, list[float] | None] = {}

        if self.electronic_steps is None:
            ionic_step_keys = ["energy", "e_fr_energy", "e_wo_entrp", "forces"]
        else:
            ionic_step_keys = ["forces"]
            flattened_e_steps = []
            for e_step in self.electronic_steps:
                flattened_e_steps.extend(e_step)

            for k, remap in {
                "e_0_energy": "energy",
                "e_fr_energy": "e_fr_energy",
                "e_wo_entrp": "e_wo_entrp",
            }.items():
                list_data = [getattr(e_step, k, None) for e_step in flattened_e_steps]
                if not all(list_data):
                    ionic_step_keys.append(remap)
                    continue

                data = np.array(list_data)
                conv_data[remap] = np.abs(data[-1] - data).tolist()

        for k in ionic_step_keys:
            if (list_data := getattr(self, k)) is None:
                conv_data[k] = None
                continue

            if k == "forces":
                list_data = [np.linalg.norm(f) for f in list_data]

            data = np.array(list_data)
            conv_data[k] = np.abs(data[-1] - data).tolist()

        return conv_data

    @classmethod
    def _calculation_to_props_dict(
        cls,
        calc: Calculation,
    ) -> tuple[dict[str, list[Any]], RunType, TaskType, CalcType]:
        """Convert a single VASP calculation to Trajectory._from_dict compatible dict.

        Parameters
        -----------
        calc (emmet.core.vasp.calculation.Calculation)

        Returns
        -----------
        dict, RunType, TaskType, CalcType
        """

        ct: CalcType | None = None
        rt: RunType | None = None
        tt: TaskType | None = None

        ionic_step_props = set(
            cls.model_fields["ionic_step_properties"].default
        ).difference(
            {
                "energy",
                "magmoms",
            }
        )
        # refresh calc, run, and task type if possible
        if calc.input:
            vis = calc.input.model_dump()
            padded_params = {
                **(calc.input.parameters or {}),
                **(calc.input.incar or {}),
            }
            ct = calc_type(vis, padded_params)
            rt = run_type(padded_params)
            tt = task_type(vis)

        props = defaultdict(list)

        for ionic_step in calc.output.ionic_steps:
            props["structure"].append(ionic_step.structure)

            props["energy"].append(ionic_step.e_0_energy)
            for k in ionic_step_props:
                props[k].append(getattr(ionic_step, k))

        return dict(props), rt, tt, ct

    @classmethod
    def from_task_doc(cls, task_doc, **kwargs) -> list["Trajectory"]:
        """
        Create trajectories from a TaskDoc.

        This class creates a separate trajectory for all non-continuous
        calculations in `TaskDoc.calcs_reversed`.
        This is to ensure that each trajectory represents only a single
        calculation type.
        In cases where sequential calculations in the reversed `calcs_reversed`
        have the same calc type, their trajectories are joined.

        Note that if no input is provided in the calculation, the calculation
        is split off into its own trajectory.

        Parameters
        -----------
        task_doc : emmet.core.TaskDoc
        kwargs
            Other kwargs to pass to Trajectory

        Returns
        -----------
        list of Trajectory
        """

        trajs: list[Trajectory] = []
        kwargs.update(
            identifier=str(task_doc.task_id) if task_doc.task_id else None,
        )

        props: dict[str, list[Any]] = {}
        old_meta = {f"{k}_type": None for k in ("run", "task", "calc")}
        new_meta = deepcopy(old_meta)
        # un-reverse the calcs_reversed
        for icr, cr in enumerate(task_doc.calcs_reversed[::-1]):
            (
                new_props,
                new_meta["run_type"],
                new_meta["task_type"],
                new_meta["calc_type"],
            ) = cls._calculation_to_props_dict(cr)

            if (
                old_meta["calc_type"] != new_meta["calc_type"]
                or new_meta["calc_type"] is None
                or icr == 0
            ):
                # Either CalcType changed or no calculation input was provided
                # or this is the first calculation in `calcs_reversed`
                # Append existing trajectory to list of trajectories, and restart
                if icr > 0:
                    trajs.append(cls._from_dict(props, **old_meta, **kwargs))  # type: ignore[arg-type]

                props = deepcopy(new_props)
            else:
                for k, new_vals in new_props.items():
                    props[k].extend(new_vals)

            for k, v in new_meta.items():
                old_meta[k] = v

        # create final trajectory
        trajs.append(cls._from_dict(props, **old_meta, **kwargs))  # type: ignore[arg-type]

        return trajs

    @requires(
        pa is not None, message="pyarrow must be installed to de-/serialize to parquet"
    )
    def to_arrow(
        self,
        file_name: str | Path | None = None,
        store_conv_data: bool = True,
        **write_file_kwargs,
    ) -> ArrowTable:
        """
        Create a PyArrow Table from a Trajectory.

        Parameters
        -----------
        file_name : str, .Path, or None (default)
            If not None, a file to write the parquet-format output to.
            Accepts any compression extension used by pyarrow.write_table
        store_conv_data : bool = True (default)
            Whether to store the data in `Trajectory.convergence_data`.
            Defaults to True to ensure that MP website convergence data is
            stored and easily accesible via parquet.
        **write_file_kwargs
            If file_name is not None, any kwargs to pass to
            pyarrow.parquet.write_file

        Returns
        -----------
        pyarrow.Table
        """
        pa_config = {
            k: pa.array([v])
            for k, v in self.model_dump(mode="json").items()
            if hasattr(v, "__len__")
        }

        for k in (
            "run_type",
            "task_type",
        ):
            if val := getattr(self, k):
                pa_config[k] = pa.array([val.value])

        if store_conv_data:
            for k, v in self.convergence_data.items():
                pa_config[f"{k}_convergence"] = pa.array([v])

        pa_config["has_full_output"] = [self.has_full_output]

        pa_table = pa.table(pa_config)
        if file_name:
            pa_pq.write_table(pa_table, file_name, **write_file_kwargs)

        return pa_table


class Trajectory(RelaxTrajectory, _MDMixin):
    """Trajectory with flexibility for electronic structure molecular dynamics."""



"""Schemas for OpenMM tasks."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd  # type: ignore[import-untyped]
from pydantic import BaseModel, ConfigDict, Field

from emmet.core.openff import MDTaskDocument  # type: ignore[import-untyped]
from emmet.core.openff.tasks import CompressedStr  # type: ignore[import-untyped]
from emmet.core.vasp.task_valid import TaskState

if TYPE_CHECKING:
    from typing import Any

from emmet.core.utils import arrow_incompatible, type_override


@arrow_incompatible
class CalculationInput(BaseModel):  # type: ignore[call-arg]
    """OpenMM input settings for a job, these are the attributes of the OpenMMMaker."""

    n_steps: int | None = Field(
        None, description="The number of simulation steps to run."
    )

    step_size: float | None = Field(
        None, description="The size of each simulation step (picoseconds)."
    )

    temperature: float | None = Field(
        None, description="The simulation temperature (kelvin)."
    )

    pressure: float | None = Field(
        None, description="The simulation pressure (atmospheres)."
    )

    friction_coefficient: float | None = Field(
        None,
        description=(
            "The friction coefficient for the integrator (inverse picoseconds)."
        ),
    )

    platform_name: str | None = Field(
        None,
        description=(
            "The name of the OpenMM platform to use, passed to "
            "Interchange.to_openmm_simulation."
        ),
    )

    platform_properties: dict | None = Field(
        None,
        description=(
            "Properties for the OpenMM platform, passed to "
            "Interchange.to_openmm_simulation."
        ),
    )

    state_interval: int | None = Field(
        None,
        description=(
            "State is saved every `state_interval` timesteps. For no state, set to 0."
        ),
    )

    state_file_name: str | None = Field(
        None, description="The name of the state file to save."
    )

    traj_interval: int | None = Field(
        None,
        description=(
            "The trajectory is saved every `traj_interval` timesteps. For no trajectory, set to 0."
        ),
    )

    wrap_traj: bool | None = Field(
        None, description="Whether to wrap trajectory coordinates."
    )

    report_velocities: bool | None = Field(
        None, description="Whether to report velocities in the trajectory file."
    )

    traj_file_name: str | None = Field(
        None, description="The name of the trajectory file to save."
    )

    traj_file_type: str | None = Field(
        None,
        description="The type of trajectory file to save.",
    )

    embed_traj: bool | None = Field(
        None,
        description="Whether to embed the trajectory blob in CalculationOutput.",
    )

    model_config = ConfigDict(extra="allow")


@type_override({"traj_blob": str})
class CalculationOutput(BaseModel):
    """OpenMM calculation output files and extracted data."""

    dir_name: str | None = Field(None, description="The directory for this OpenMM task")

    traj_file: str | None = Field(
        None, description="Path to the trajectory file relative to `dir_name`"
    )

    traj_blob: CompressedStr | None = Field(
        None, description="Trajectory file bytes blob hex encoded to a string"
    )

    state_file: str | None = Field(
        None, description="Path to the state file relative to `dir_name`"
    )

    steps_reported: list[int] | None = Field(
        None, description="Steps where outputs are reported"
    )

    time: list[float] | None = Field(None, description="List of times")

    potential_energy: list[float] | None = Field(
        None, description="List of potential energies"
    )

    kinetic_energy: list[float] | None = Field(
        None, description="List of kinetic energies"
    )

    total_energy: list[float] | None = Field(None, description="List of total energies")

    temperature: list[float] | None = Field(None, description="List of temperatures")

    volume: list[float] | None = Field(None, description="List of volumes")

    density: list[float] | None = Field(None, description="List of densities")

    elapsed_time: float | None = Field(
        None, description="Elapsed time for the calculation (seconds)."
    )

    @classmethod
    def from_directory(
        cls,
        dir_name: Path | str,
        state_file_name: str,
        traj_file_name: str,
        elapsed_time: float | None = None,
        embed_traj: bool = False,
    ) -> CalculationOutput:
        """Extract data from the output files in the directory."""
        state_file = Path(dir_name) / state_file_name
        column_name_map: dict[str, str] = {
            '#"Step"': "steps_reported",
            "Potential Energy (kJ/mole)": "potential_energy",
            "Kinetic Energy (kJ/mole)": "kinetic_energy",
            "Total Energy (kJ/mole)": "total_energy",
            "Temperature (K)": "temperature",
            "Box Volume (nm^3)": "volume",
            "Density (g/mL)": "density",
        }
        state_is_not_empty = state_file.exists() and state_file.stat().st_size > 0
        if state_is_not_empty:
            data = pd.read_csv(state_file, header=0)
            data = data.rename(columns=column_name_map)
            data = data.filter(items=list(column_name_map.values()))
            attributes: dict[str, Any] = data.to_dict(orient="list")  # type: ignore[assignment]
        else:
            attributes = {name: None for name in column_name_map.values()}
            state_file_name = None  # type: ignore[assignment]

        traj_file = Path(dir_name) / traj_file_name
        traj_is_not_empty = traj_file.exists() and traj_file.stat().st_size > 0
        traj_file_name = traj_file_name if traj_is_not_empty else None  # type: ignore

        traj_blob: str | None = None
        if traj_is_not_empty and embed_traj:
            with open(traj_file, "rb") as f:
                traj_blob = f.read().hex()

        return CalculationOutput(
            dir_name=str(dir_name),
            elapsed_time=elapsed_time,
            traj_file=traj_file_name,
            state_file=state_file_name,
            traj_blob=traj_blob,
            **attributes,
        )


@arrow_incompatible
class Calculation(BaseModel):
    """All input and output data for an OpenMM calculation."""

    dir_name: str | None = Field(
        None, description="The directory for this OpenMM calculation"
    )

    has_openmm_completed: TaskState | bool | None = Field(
        None, description="Whether OpenMM completed the calculation successfully"
    )

    input: CalculationInput | None = Field(
        None, description="OpenMM input settings for the calculation"
    )
    output: CalculationOutput | None = Field(
        None, description="The OpenMM calculation output"
    )

    completed_at: str | None = Field(
        None, description="Timestamp for when the calculation was completed"
    )
    task_name: str | None = Field(
        None, description="Name of task given by custodian (e.g., relax1, relax2)"
    )

    calc_type: str | None = Field(
        None,
        description="Return calculation type (run type + task_type). or just new thing",
    )


class OpenMMTaskDocument(MDTaskDocument):
    """Definition of the OpenMM task document."""

    calcs_reversed: list[Calculation] | None = Field(
        None,
        title="Calcs reversed data",
        description="Detailed data for each OpenMM calculation contributing to the "
        "task document.",
    )



import pytest
from pymatgen.core.structure import Structure

from . import test_structures
from emmet.core.utils import utcnow

try:
    from emmet.core.optimade import OptimadeMaterialsDoc
except Exception:
    pytest.skip("could not import 'optimade' ", allow_module_level=True)


@pytest.mark.xfail(reason="Optimade + fastapi issues.")
@pytest.mark.parametrize("structure", test_structures.values())
def test_oxidation_state(structure: Structure):
    """Very simple test to make sure this actually works"""
    print(f"Should work : {structure.composition}")
    doc = OptimadeMaterialsDoc.from_structure(
        structure, material_id=33, last_updated=utcnow()
    )
    assert doc is not None



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import BaseModel, Field
from pymatgen.core import Structure
from pymatgen.core.structure import Molecule
from pymatgen.symmetry.analyzer import (
    PointGroupAnalyzer,
    SpacegroupAnalyzer,
    SymmetryUndeterminedError,
    spglib,
)

from emmet.core.settings import EmmetSettings
from emmet.core.types.enums import ValueEnum

if TYPE_CHECKING:
    from typing import Any

SETTINGS = EmmetSettings()


class CrystalSystem(ValueEnum):
    """
    The crystal system of the lattice
    """

    tri = "Triclinic"
    mono = "Monoclinic"
    ortho = "Orthorhombic"
    tet = "Tetragonal"
    trig = "Trigonal"
    hex_ = "Hexagonal"
    cubic = "Cubic"


class PointGroupData(BaseModel):
    """
    Defines symmetry for a molecule document
    """

    point_group: str | None = Field(
        None, title="Point Group Symbol", description="The point group for the lattice"
    )

    rotation_number: float | None = Field(
        None,
        title="Rotational Symmetry Number",
        description="Rotational symmetry number for the molecule",
    )

    linear: bool | None = Field(
        None, title="Molecule Linearity", description="Is the molecule linear?"
    )

    tolerance: float | None = Field(
        None,
        title="Point Group Analyzer Tolerance",
        description="Distance tolerance to consider sites as symmetrically equivalent.",
    )

    eigen_tolerance: float | None = Field(
        None,
        title="Interia Tensor Eigenvalue Tolerance",
        description="Tolerance to compare eigen values of the inertia tensor.",
    )

    matrix_tolerance: float | None = Field(
        None,
        title="Symmetry Operation Matrix Element Tolerance",
        description="Tolerance used to generate the full set of symmetry operations of the point group.",
    )

    @classmethod
    def from_molecule(cls, molecule: Molecule) -> "PointGroupData":
        tol = SETTINGS.PGATOL
        eigentol = SETTINGS.PGAEIGENTOL
        matrixtol = SETTINGS.PGAMATRIXTOL
        pga = PointGroupAnalyzer(
            molecule,
            tolerance=tol,
            eigen_tolerance=eigentol,
            matrix_tolerance=matrixtol,
        )
        symmetry: dict[str, Any] = {
            "tolerance": tol,
            "eigen_tolerance": eigentol,
            "matrix_tolerance": matrixtol,
            "point_group": pga.sch_symbol,
        }

        rotational_symmetry_numbers = {
            1.0: ["C1", "Cs", "Ci", "C*v", "S2"],
            2.0: ["C2", "C2h", "C2v", "S4", "D*h"],
            3.0: ["C3", "C3h", "C3v", "S6"],
            4.0: ["C4v", "D4h", "D4d", "D2", "D2h", "D2d"],
            5.0: ["C5v", "Ih"],
            6.0: ["D3", "D3h", "D3d"],
            10.0: ["D5h", "D5d"],
            12.0: ["T", "Td", "Th", "D6h"],
            14.0: ["D7h"],
            16.0: ["D8h"],
            24.0: ["Oh"],
            float("inf"): ["Kh"],
        }

        r = 1.0
        for rot_num, point_groups in rotational_symmetry_numbers.items():
            if symmetry["point_group"] in point_groups:
                r = rot_num
                break
        if symmetry["point_group"] in ["C*v", "D*h"]:
            linear = True
        else:
            linear = False

        symmetry["rotation_number"] = float(r)
        symmetry["linear"] = linear

        return PointGroupData(**symmetry)


class SymmetryData(BaseModel):
    """
    Defines a symmetry data set for materials documents
    """

    crystal_system: CrystalSystem | None = Field(
        None, title="Crystal System", description="The crystal system for this lattice."
    )

    symbol: str | None = Field(
        None,
        title="Space Group Symbol",
        description="The spacegroup symbol for the lattice.",
    )

    hall: str | None = Field(
        None,
        title="Hall Symbol",
        description="Hall symbol for the lattice",
    )

    number: int | None = Field(
        None,
        title="Space Group Number",
        description="The spacegroup number for the lattice.",
    )

    point_group: str | None = Field(
        None, title="Point Group Symbol", description="The point group for the lattice."
    )

    symprec: float | None = Field(
        None,
        title="Symmetry Finding Precision",
        description="The precision provided to spglib to determine the symmetry of this structure.",
    )

    angle_tolerance: float | None = Field(
        None,
        title="Angle Tolerance",
        description="Angle tolerance provided to spglib to determine the symmetry of this structure.",
    )

    version: str | None = Field(None, title="spglib version")

    @classmethod
    def from_structure(cls, structure: Structure) -> "SymmetryData":
        symmetry: dict[str, Any] = {
            "symbol": None,
            "number": None,
            "point_group": None,
            "crystal_system": None,
            "hall": None,
            "version": spglib.__version__,
            "symprec": SETTINGS.SYMPREC,
            "angle_tolerance": SETTINGS.ANGLE_TOL,
        }

        try:
            sg = SpacegroupAnalyzer(
                structure,
                symprec=symmetry["symprec"],
                angle_tolerance=symmetry["angle_tolerance"],
            )
        except SymmetryUndeterminedError:
            try:
                symmetry["symprec"] = 1e-3
                symmetry["angle_tolerance"] = 1
                sg = SpacegroupAnalyzer(
                    structure,
                    symprec=symmetry["symprec"],
                    angle_tolerance=symmetry["angle_tolerance"],
                )
            except SymmetryUndeterminedError:
                return SymmetryData(**symmetry)

        symmetry.update(
            {
                "symbol": sg.get_space_group_symbol(),
                "number": sg.get_space_group_number(),
                "point_group": sg.get_point_group_symbol(),
                "crystal_system": CrystalSystem(sg.get_crystal_system().title()),
                "hall": sg.get_hall(),
            }
        )

        return SymmetryData(**symmetry)



"""Test extended MPID / AlphaID formats."""

from itertools import product
import numpy as np
from pymatgen.core import Element

from emmet.core.mpid import MPID, AlphaID
from emmet.core.mpid_ext import BatteryID, ThermoID, XasSpectrumID, validate_identifier
from emmet.core.types.enums import ThermoType, XasEdge, XasType


def random_identifiers(num_idx: int = 10) -> set[str | MPID | AlphaID]:
    randints = set()
    c = 0
    while len(randints) < num_idx:
        # ensure mix of IDs which are MPID and AlphaID
        if c % 2 == 0:
            low = 1
            high = AlphaID._cut_point
        else:
            low = AlphaID._cut_point
            high = 1_000_000_000
        randints.add(np.random.randint(low=low, high=high))
        c += 1

    idxs = set()
    for idx in randints:
        idxs.update({f"mp-{idx}", MPID(f"mvc-{idx}"), AlphaID(idx)})
    return idxs


RAND_IDS = random_identifiers()


def test_validate_on_mpid_alpha():

    for idx in RAND_IDS:
        assert validate_identifier(idx, serialize=False) == AlphaID(idx).formatted
        assert validate_identifier(idx, serialize=True) == str(AlphaID(idx))


def test_single_suffix_ids():

    last_iden = None
    for id_cls, enum_cls in {BatteryID: Element, ThermoID: ThermoType}.items():

        for idx, enum_val in product(RAND_IDS, enum_cls.__members__):
            iden = id_cls(identifier=idx, suffix=enum_val)
            assert iden == id_cls.from_str(f"{idx}_{enum_val}")
            if last_iden:
                assert iden != last_iden
            last_iden = iden


def test_many_suffix_id():

    last_iden = None
    for prod in product(
        RAND_IDS, XasType.__members__, Element.__members__, XasEdge.__members__
    ):
        idx, xas_type, ele, xas_edge = prod
        iden = XasSpectrumID(identifier=idx, suffix=(xas_type, ele, xas_edge))
        assert iden == XasSpectrumID.from_str("-".join(prod))
        if last_iden:
            assert iden != last_iden
        last_iden = iden



"""Define schemas for ML calculations and training data."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from pydantic import (
    BaseModel,
    Field,
    field_validator,
    model_serializer,
    model_validator,
)
from pymatgen.analysis.elasticity import ElasticTensor
from pymatgen.core import Element, Structure

from emmet.core.elasticity import (
    BulkModulus,
    ElasticityDoc,
    ElasticTensorDoc,
    ShearModulus,
)
from emmet.core.math import Matrix3D, Vector3D, Vector6D, matrix_3x3_to_voigt
from emmet.core.structure import StructureMetadata
from emmet.core.tasks import TaskDoc
from emmet.core.types.pymatgen_types.composition_adapter import CompositionType
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.types.typing import IdentifierType
from emmet.core.utils import jsanitize
from emmet.core.vasp.calc_types import RunType as VaspRunType

if TYPE_CHECKING:
    from collections.abc import Sequence

    from typing_extensions import Self


class MLDoc(ElasticityDoc):
    """Schema for machine learning interatomic potential calculations.

    Attributes:
    - metadata
        - material_id (str): MP ID
        - structure (Structure): pymatgen Structure object
        - deprecated (bool): whether this material is deprecated in MP
        - calculator (str): name of model used as ML potential.
        - version (str): version of matcalc used to generate this document
    - relaxation
        - final_structure: relaxed pymatgen Structure object
        - energy (float): final energy in eV
        - volume (float): final volume in Angstrom^3
        - lattice parameters (float): a, b, c, alpha, beta, gamma
    - equation of state
        - eos (dict[str, list[float]]): with keys energies and volumes
        - bulk_modulus_bm (float): Birch-Murnaghan bulk modulus in GPa
    - phonon
        - temperatures (list[float]): temperatures in K
        - free_energy (list[float]): Helmholtz energies at those temperatures in eV
        - entropy (list[float]): entropies at those temperatures in eV/K
        - heat_capacities (list[float]): heat capacities at constant volume in eV/K
    - elasticity
        - elastic_tensor (ElasticTensorDoc): pydantic model from emmet.core.elasticity
        - shear_modulus (ShearModulus): Voigt-Reuss-Hill shear modulus (single float in eV/Angstrom^3)
        - bulk_modulus (BulkModulus): Voigt-Reuss-Hill bulk modulus (single float in eV/Angstrom^3)
        - youngs_modulus (float): Young's modulus (single float in eV/Angstrom^3)
    """

    property_name: str = "ml"

    # metadata
    structure: StructureType = Field(description="Original structure")
    model: str | None = Field(None, description="Name of model used as ML potential.")
    version: str | None = Field(
        None, description="Version of model used as ML potential"
    )

    # relaxation attributes
    final_structure: StructureType | None = Field(
        None, description="ML-potential-relaxed structure"
    )
    energy: float | None = Field(None, description="Final energy in eV")
    volume: float | None = Field(None, description="Final volume in Angstrom^3")
    a: float | None = Field(None, description="Lattice length a in Angstrom")
    b: float | None = Field(None, description="Lattice length b in Angstrom")
    c: float | None = Field(None, description="Lattice length c in Angstrom")
    alpha: float | None = Field(None, description="Lattice angle alpha in degrees")
    beta: float | None = Field(None, description="Lattice angle beta in degrees")
    gamma: float | None = Field(None, description="Lattice angle gamma in degrees")

    # equation of state attributes
    eos: dict[str, list[float]] | None = Field(
        None, description="dict with keys energies and volumes"
    )
    bulk_modulus_bm: float | None = Field(None, description="bm.b0_GPa")

    # phonons attributes
    temperatures: list[float] | None = Field(None, description="list of temperatures")
    free_energy: list[float] | None = Field(
        None,
        description="list of Helmholtz free energies at corresponding temperatures",
    )
    entropy: list[float] | None = Field(
        None, description="list of entropies at corresponding temperatures in eV/K"
    )
    heat_capacity: list[float] | None = Field(
        None,
        description="list of heat capacities at constant volume at corresponding "
        "temperatures in eV/K",
    )

    # elasticity attributes
    # all inherited from ElasticityDoc

    @field_validator("elastic_tensor", mode="before")
    def elastic_tensor(cls, val) -> ElasticTensorDoc:
        """ElasticTensorDoc from MSONable dict of ElasticTensor, or list (specifying the Voigt array)
        or the ElasticTensor class itself.
        """
        if isinstance(val, dict):
            tensor = ElasticTensor.from_dict(val)
        elif isinstance(val, (list, tuple)):
            tensor = ElasticTensor(val)
        else:
            tensor = val
        return ElasticTensorDoc(raw=tensor.voigt.tolist())

    @field_validator("bulk_modulus", mode="before")
    def bulk_vrh_no_suffix(cls, new_key, values):
        """Map field bulk_modulus_vrh to bulk_modulus."""
        val = values.get("bulk_modulus_vrh", new_key)
        return BulkModulus(vrh=val)

    @field_validator("shear_modulus", mode="before")
    def shear_vrh_no_suffix(cls, new_key, values):
        """Map field shear_modulus_vrh to shear_modulus."""
        val = values.get("shear_modulus_vrh", new_key)
        return ShearModulus(vrh=val)


class MLTrainDoc(StructureMetadata):
    """Generic schema for ML training data."""

    structure: StructureType | None = Field(
        None, description="Structure for this entry."
    )

    energy: float | None = Field(
        None, description="The total energy associated with this structure."
    )

    forces: list[Vector3D] | None = Field(
        None,
        description="The interatomic forces corresponding to each site in the structure.",
    )

    abs_forces: list[float] | None = Field(
        None, description="The magnitude of the interatomic force on each site."
    )

    stress: Vector6D | None = Field(
        None,
        description="The components of the symmetric stress tensor in Voigt notation (xx, yy, zz, yz, xz, xy).",
    )

    stress_matrix: Matrix3D | None = Field(
        None,
        description="The 3x3 stress tensor. Use this if the tensor is unphysically non-symmetric.",
    )

    bandgap: float | None = Field(None, description="The final DFT bandgap.")

    elements: list[Element] | None = Field(
        None,
        description="List of unique elements in the material sorted alphabetically.",
    )

    composition: CompositionType | None = Field(
        None, description="Full composition for the material."
    )

    composition_reduced: CompositionType | None = Field(
        None,
        title="Reduced Composition",
        description="Simplified representation of the composition.",
    )

    functional: VaspRunType | None = Field(
        None, description="The approximate functional used to generate this entry."
    )

    bader_charges: list[float] | None = Field(
        None, description="Bader charges on each site of the structure."
    )
    bader_magmoms: list[float] | None = Field(
        None,
        description="Bader on-site magnetic moments for each site of the structure.",
    )

    @model_serializer
    def deseralize(self):
        """Ensure output is JSON compliant."""
        return jsanitize({k: getattr(self, k, None) for k in self.model_fields})

    @model_validator(mode="after")
    def set_abs_forces(
        self,
    ) -> Self:
        """Ensure the abs_forces are set if the vector-valued forces are."""
        if self.forces is not None and self.abs_forces is None:
            self.abs_forces = [np.linalg.norm(f) for f in self.forces]  # type: ignore[misc]
        return self

    @classmethod
    def from_structure(
        cls,
        meta_structure: Structure,
        fields: list[str] | None = None,
        **kwargs,
    ) -> Self:
        """
        Create an ML training document from a structure and fields.

        This method mostly exists to ensure that the structure field is
        set because `meta_structure` does not populate it automatically.

        Parameters
        -----------
        meta_structure : Structure
        fields : list of str or None
            Additional fields in the document to populate
        **kwargs
            Any other fields / constructor kwargs
        """
        return super().from_structure(
            meta_structure=meta_structure,
            fields=fields,
            structure=meta_structure,
            **kwargs,
        )

    @classmethod
    def from_task_doc(
        cls,
        task_doc: TaskDoc,
        **kwargs,
    ) -> list[Self]:
        """Create a list of ML training documents from the ionic steps in a TaskDoc.

        Parameters
        -----------
        task_doc : TaskDoc
        **kwargs
            Any kwargs to pass to `from_structure`.
        """
        entries = []

        for cr in task_doc.calcs_reversed[::-1]:
            nion = len(cr.output.ionic_steps)

            for iion, ionic_step in enumerate(cr.output.ionic_steps):
                structure = Structure.from_dict(ionic_step.structure.as_dict())
                # these are fields that should only be set on the final frame of a calculation
                # also patch in magmoms because of how Calculation works
                last_step_kwargs = {}
                if iion == nion - 1:
                    if magmom := cr.output.structure.site_properties.get("magmom"):
                        structure.add_site_property("magmom", magmom)
                    last_step_kwargs["bandgap"] = cr.output.bandgap
                    if bader_analysis := cr.bader:
                        for bk in (
                            "charge",
                            "magmom",
                        ):
                            last_step_kwargs[f"bader_{bk}s"] = bader_analysis[bk]

                if (_st := ionic_step.stress) is not None:
                    st = np.array(_st)
                    if np.allclose(st, st.T, rtol=1e-8):
                        # Stress tensor is symmetric
                        last_step_kwargs["stress"] = matrix_3x3_to_voigt(_st)
                    else:
                        # Stress tensor is non-symmetric
                        last_step_kwargs["stress_matrix"] = _st

                entries.append(
                    cls.from_structure(
                        meta_structure=structure,
                        energy=ionic_step.e_0_energy,
                        forces=ionic_step.forces,
                        functional=cr.run_type,
                        **last_step_kwargs,
                        **kwargs,
                    )
                )
        return entries


class MatPESProvenanceDoc(BaseModel):
    """Information regarding the origins of a MatPES structure."""

    original_mp_id: IdentifierType | None = Field(
        None,
        description="MP identifier corresponding to the Materials Project structure from which this entry was sourced from.",
    )
    materials_project_version: str | None = Field(
        None,
        description="The version of the Materials Project from which the struture was sourced.",
    )
    md_ensemble: str | None = Field(
        None,
        description="The molecular dynamics ensemble used to generate this structure.",
    )
    md_temperature: float | None = Field(
        None,
        description="If a float, the temperature in Kelvin at which MLMD was performed.",
    )
    md_pressure: float | None = Field(
        None,
        description="If a float, the pressure in atmosphere at which MLMD was performed.",
    )
    md_step: int | None = Field(
        None,
        description="The step in the MD simulation from which the structure was sampled.",
    )
    mlip_name: str | None = Field(
        None, description="The name of the ML potential used to perform MLMD."
    )


class MatPESTrainDoc(MLTrainDoc):
    """
    Schema for VASP data in the Materials Potential Energy Surface (MatPES) effort.

    This schema is used in the data entries for MatPES v2025.1,
    which can be downloaded either:
        - On [MPContribs](https://materialsproject-contribs.s3.amazonaws.com/index.html#MatPES_2025_1/)
        - or on [the site]
    """

    matpes_id: str | None = Field(None, description="MatPES identifier.")

    formation_energy_per_atom: float | None = Field(
        None,
        description="The uncorrected formation enthalpy per atom at zero pressure and temperature.",
    )
    cohesive_energy_per_atom: float | None = Field(
        None, description="The uncorrected cohesive energy per atom."
    )

    provenance: MatPESProvenanceDoc | None = Field(
        None, description="Information about the provenance of the structure."
    )

    @property
    def pressure(self) -> float | None:
        """Return the pressure from the DFT stress tensor."""
        return sum(self.stress[:3]) / 3.0 if self.stress else None

    @property
    def magmoms(self) -> Sequence[float] | None:
        """Retrieve on-site magnetic moments from the structure if they exist."""
        magmom = (
            self.structure.site_properties.get("magmom") if self.structure else None
        )
        return magmom



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.entries.mixing_scheme import MaterialsProjectDFTMixingScheme
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.compatibility_adapter import (
    TypedCompatibilityDict,
    TypedMaterialsProject2020CompatibilityAdapterDict,
)
from emmet.core.types.pymatgen_types.structure_matcher_adapter import (
    TypedStructureMatcherDict,
)

TypedMaterialsProjectDFTMixingSchemeDict = TypedDict(
    "TypedMaterialsProjectDFTMixingSchemeDict",
    {
        "@module": str,
        "@class": str,
        "@version": str,
        "structure_matcher": TypedStructureMatcherDict,
        "run_type_1": str,
        "run_type_2": str,
        "compat_1": TypedMaterialsProject2020CompatibilityAdapterDict,
        "compat_2": TypedCompatibilityDict,
        "fuzzy_matching": bool,
        "check_potcar": bool,
    },
)

MaterialsProjectDFTMixingSchemeTypeVar = TypeVar(
    "MaterialsProjectDFTMixingSchemeTypeVar",
    MaterialsProjectDFTMixingScheme,
    TypedMaterialsProjectDFTMixingSchemeDict,
)

MaterialsProjectDFTMixingSchemeType = Annotated[
    MaterialsProjectDFTMixingSchemeTypeVar,
    BeforeValidator(
        lambda x: (
            MaterialsProjectDFTMixingScheme.from_dict(x) if isinstance(x, dict) else x
        )
    ),
    WrapSerializer(
        lambda x, nxt, info: x.as_dict(),
        return_type=TypedMaterialsProjectDFTMixingSchemeDict,
    ),
]



from typing import Annotated, TypeVar

from monty.json import MontyDecoder
from pydantic import BeforeValidator, WrapSerializer
from pymatgen.analysis.defects.core import Defect
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.sites_adapter import MSONableTypedSiteDict
from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedStructureDict,
    pop_empty_structure_keys,
)

TypedDefectDict = TypedDict(
    "TypedDefectDict",
    {
        "@module": str,
        "@class": str,
        "@version": str,
        "structure": TypedStructureDict,
        "site": MSONableTypedSiteDict,
        "multiplicity": int,
        "oxi_state": float,
        "equivalent_sites": list[MSONableTypedSiteDict],
        "symprec": float,
        "angle_tolerance": float,
        "user_changes": list[int],
    },
)

DefectTypeVar = TypeVar("DefectTypeVar", Defect, TypedDefectDict)


def pop_defect_empty_structure_fields(defect: DefectTypeVar):
    if isinstance(defect, dict):
        defect["structure"] = pop_empty_structure_keys(defect["structure"])
        return MontyDecoder().process_decoded(defect)

    return defect


DefectType = Annotated[
    DefectTypeVar,
    BeforeValidator(pop_defect_empty_structure_fields),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedDefectDict),
]



"""Module defining VASP calculation types."""

from emmet.core.vasp.calc_types.enums import CalcType, RunType, TaskType
from emmet.core.vasp.calc_types.utils import calc_type, run_type, task_type

__all__ = ["CalcType", "RunType", "TaskType", "calc_type", "run_type", "task_type"]



"""Test basic features of XASDoc."""

import pytest
from monty.serialization import loadfn
from pymatgen.analysis.xas.spectrum import XAS
from pymatgen.core import Element

from emmet.core import ARROW_COMPATIBLE
from emmet.core.utils import jsanitize
from emmet.core.xas import XASDoc

if ARROW_COMPATIBLE:
    import pyarrow as pa

    from emmet.core.arrow import arrowize


@pytest.fixture(scope="module")
def xas_dict(test_dir):
    return loadfn(test_dir / "xasdoc_nonpos_mp_626735.json.gz", cls=None)


def test_xas_doc(xas_dict):
    # First show that there are non-positive intensities
    non_pos_idx = [idx for idx, v in enumerate(xas_dict["spectrum"]["y"]) if v <= 0.0]
    assert len(non_pos_idx) > 0

    # Now show that XASDoc removes non-positive intensities and correctly serializes
    xas = XASDoc(**xas_dict)
    assert isinstance(xas.spectrum, XAS)
    assert len(xas.spectrum.y[xas.spectrum.y <= 0.0]) == 0
    assert all(
        xas.spectrum.y[idx] > 0.0 and xas.spectrum.y[idx] == pytest.approx(0.0)
        for idx in non_pos_idx
    )

    assert isinstance(xas.absorbing_element, Element)


@pytest.mark.skipif(
    not ARROW_COMPATIBLE, reason="pyarrow must be installed to run this test."
)
def test_arrow(xas_dict):
    doc = XASDoc(**xas_dict)
    arrow_struct = pa.scalar(
        doc.model_dump(context={"format": "arrow"}), type=arrowize(XASDoc)
    )
    test_arrow_doc = XASDoc(**arrow_struct.as_py(maps_as_pydicts="strict"))

    assert jsanitize(doc.model_dump(), allow_bson=True) == jsanitize(
        test_arrow_doc.model_dump(), allow_bson=True
    )



from pydantic import Field
from pydantic.main import BaseModel

from emmet.core.types.enums import ValueEnum
from emmet.core.types.pymatgen_types.structure_adapter import StructureType


class MPCompleteDoc(BaseModel):
    """
    Defines data for MPComplete structure submissions
    """

    structure: StructureType | None = Field(
        None,
        description="Structure submitted by the user.",
    )

    public_name: str | None = Field(
        None,
        description="Public name of submitter.",
    )

    public_email: str | None = Field(
        None,
        description="Public email of submitter.",
    )


class MPCompleteDataStatus(ValueEnum):
    """
    Submission status for MPComplete data
    """

    submitted = "SUBMITTED"
    pending = "PENDING"
    running = "RUNNING"
    error = "ERROR"
    complete = "COMPLETE"



import json
import copy
import datetime

import pytest

import numpy as np

from monty.io import zopen

from pymatgen.core.structure import Molecule

from emmet.core.qchem.task import TaskDocument
from emmet.core.molecules.trajectory import ForcesDoc, TrajectoryDoc


@pytest.fixture(scope="session")
def force_calc(test_dir):
    with zopen(test_dir / "force_trajectory" / "force_calc.json.gz") as f:
        data = json.load(f)

    data["last_updated"] = datetime.datetime.now()

    task = TaskDocument(**data)
    return task


@pytest.fixture(scope="session")
def ffopt_task(test_dir):
    with zopen(test_dir / "force_trajectory" / "ffopt_calc.json.gz") as f:
        data = json.load(f)

    data["last_updated"] = datetime.datetime.now()

    task = TaskDocument(**data)
    return task


def test_forces(force_calc):
    doc = ForcesDoc.from_task(
        force_calc, molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )
    assert doc.property_name == "forces"
    assert len(doc.forces) == 11
    assert np.allclose(np.array(doc.forces), np.array(doc.precise_forces), atol=0.001)
    assert doc.cds_forces is None
    assert doc.average_force_magnitude == pytest.approx(0.03861184536372114)
    assert doc.max_force_magnitude == pytest.approx(0.07195037119572352)
    assert doc.min_force_magnitude == pytest.approx(0.003093974384847178)


def test_trajectory(ffopt_task):
    doc = TrajectoryDoc.from_task(
        ffopt_task, molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )

    assert doc.property_name == "optimization_trajectory"
    assert doc.num_trajectories == 2
    assert len(doc.geometries) == doc.num_trajectories
    assert len(doc.energies) == doc.num_trajectories
    assert len(doc.forces) == doc.num_trajectories

    assert doc.species == ["C", "C", "C", "O", "C", "N", "O", "N", "N", "H", "H", "H"]
    assert len(doc.geometries[0]) == 13
    assert doc.energies[0][0] == pytest.approx(-468.1895254889)

    mol_objects = doc.molecules
    assert isinstance(mol_objects[0][0], Molecule)
    assert len(mol_objects) == doc.num_trajectories

    traj_objects = doc.as_trajectories()
    assert len(traj_objects) == 2


def test_missing_trajectory(ffopt_task):
    # Mess with one of the trajectories, removing gradients
    missing_grad = copy.deepcopy(ffopt_task)

    missing_grad.calcs_reversed[-1]["gradients"] = list()

    doc = TrajectoryDoc.from_task(
        missing_grad, molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )

    # The trajectory with missing gradients should not be included
    assert doc.num_trajectories == 1

    # Likewise, if there are missing geometries
    missing_geom = copy.deepcopy(ffopt_task)

    missing_geom.calcs_reversed[-1]["geometries"] = list()

    doc = TrajectoryDoc.from_task(
        missing_geom, molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )
    assert doc.num_trajectories == 1

    # And missing energies
    missing_energies = copy.deepcopy(ffopt_task)
    missing_energies.calcs_reversed[-1]["energy_trajectory"] = None

    doc = TrajectoryDoc.from_task(
        missing_energies, molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )
    assert doc.num_trajectories == 1



"""Ensure that document models used in API do not change fields."""

import pytest
from importlib import import_module


ref_model_fields = {
    "emmet.core._general_store.GeneralStoreDoc": [
        "kind",
        "markdown",
        "meta",
        "last_updated",
    ],
    "emmet.core._messages.MessagesDoc": [
        "title",
        "body",
        "authors",
        "type",
        "last_updated",
    ],
    "emmet.core._user_settings.UserSettingsDoc": ["consumer_id", "settings"],
    "emmet.core.absorption.AbsorptionDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "task_id",
        "energies",
        "energy_max",
        "absorption_coefficient",
        "average_imaginary_dielectric",
        "average_real_dielectric",
        "bandgap",
        "nkpoints",
    ],
    "emmet.core.alloys.AlloyPairDoc": ["builder_meta", "alloy_pair", "pair_id"],
    "emmet.core.bonds.BondingDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "structure_graph",
        "method",
        "bond_types",
        "bond_length_stats",
        "coordination_envs",
        "coordination_envs_anonymous",
    ],
    "emmet.core.chemenv.ChemEnvDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "valences",
        "species",
        "chemenv_symbol",
        "chemenv_iupac",
        "chemenv_iucr",
        "chemenv_name",
        "chemenv_name_with_alternatives",
        "csm",
        "method",
        "mol_from_site_environments",
        "wyckoff_positions",
    ],
    "emmet.core.dois.DOIDoc": ["doi", "bibtex", "material_id"],
    "emmet.core.elasticity.ElasticityDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "order",
        "elastic_tensor",
        "compliance_tensor",
        "bulk_modulus",
        "shear_modulus",
        "sound_velocity",
        "thermal_conductivity",
        "young_modulus",
        "universal_anisotropy",
        "homogeneous_poisson",
        "debye_temperature",
        "fitting_data",
        "fitting_method",
        "state",
    ],
    "emmet.core.electrode.InsertionElectrodeDoc": [
        "builder_meta",
        "battery_type",
        "battery_id",
        "thermo_type",
        "battery_formula",
        "working_ion",
        "num_steps",
        "max_voltage_step",
        "last_updated",
        "framework",
        "framework_formula",
        "elements",
        "nelements",
        "chemsys",
        "formula_anonymous",
        "warnings",
        "formula_charge",
        "formula_discharge",
        "max_delta_volume",
        "average_voltage",
        "capacity_grav",
        "capacity_vol",
        "energy_grav",
        "energy_vol",
        "fracA_charge",
        "fracA_discharge",
        "stability_charge",
        "stability_discharge",
        "id_charge",
        "id_discharge",
        "host_structure",
        "adj_pairs",
        "material_ids",
        "entries_composition_summary",
        "electrode_object",
    ],
    "emmet.core.electronic_structure.ElectronicStructureDoc": [
        "task_id",
        "band_gap",
        "cbm",
        "vbm",
        "efermi",
        "is_gap_direct",
        "is_metal",
        "magnetic_ordering",
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "bandstructure",
        "dos",
    ],
    "emmet.core.eos.EOSDoc": ["energies", "volumes", "eos", "material_id"],
    "emmet.core.grain_boundary.GrainBoundaryDoc": [
        "material_id",
        "sigma",
        "type",
        "rotation_axis",
        "gb_plane",
        "rotation_angle",
        "gb_energy",
        "initial_structure",
        "final_structure",
        "pretty_formula",
        "w_sep",
        "cif",
        "chemsys",
        "last_updated",
    ],
    "emmet.core.magnetism.MagnetismDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "ordering",
        "is_magnetic",
        "exchange_symmetry",
        "num_magnetic_sites",
        "num_unique_magnetic_sites",
        "types_of_magnetic_species",
        "magmoms",
        "total_magnetization",
        "total_magnetization_normalized_vol",
        "total_magnetization_normalized_formula_units",
    ],
    "emmet.core.molecules.atomic.PartialChargesDoc": [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "property_name",
        "property_id",
        "molecule_id",
        "deprecated",
        "deprecation_reasons",
        "level_of_theory",
        "solvent",
        "lot_solvent",
        "last_updated",
        "origins",
        "warnings",
        "method",
        "partial_charges",
    ],
    "emmet.core.molecules.atomic.PartialSpinsDoc": [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "property_name",
        "property_id",
        "molecule_id",
        "deprecated",
        "deprecation_reasons",
        "level_of_theory",
        "solvent",
        "lot_solvent",
        "last_updated",
        "origins",
        "warnings",
        "method",
        "partial_spins",
    ],
    "emmet.core.molecules.bonds.MoleculeBondingDoc": [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "property_name",
        "property_id",
        "molecule_id",
        "deprecated",
        "deprecation_reasons",
        "level_of_theory",
        "solvent",
        "lot_solvent",
        "last_updated",
        "origins",
        "warnings",
        "molecule_graph",
        "method",
        "bond_types",
        "bonds",
        "bonds_nometal",
    ],
    "emmet.core.molecules.orbitals.OrbitalDoc": [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "property_name",
        "property_id",
        "molecule_id",
        "deprecated",
        "deprecation_reasons",
        "level_of_theory",
        "solvent",
        "lot_solvent",
        "last_updated",
        "origins",
        "warnings",
        "open_shell",
        "nbo_population",
        "nbo_lone_pairs",
        "nbo_bonds",
        "nbo_three_center_bonds",
        "nbo_hyperbonds",
        "nbo_interactions",
        "alpha_population",
        "beta_population",
        "alpha_lone_pairs",
        "beta_lone_pairs",
        "alpha_bonds",
        "beta_bonds",
        "alpha_three_center_bonds",
        "beta_three_center_bonds",
        "alpha_hyperbonds",
        "beta_hyperbonds",
        "alpha_interactions",
        "beta_interactions",
    ],
    "emmet.core.molecules.redox.RedoxDoc": [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "property_name",
        "property_id",
        "molecule_id",
        "deprecated",
        "deprecation_reasons",
        "level_of_theory",
        "solvent",
        "lot_solvent",
        "last_updated",
        "origins",
        "warnings",
        "base_property_id",
        "electron_affinity",
        "ea_task_id",
        "ionization_energy",
        "ie_task_id",
        "reduction_energy",
        "reduction_free_energy",
        "red_molecule_id",
        "red_property_id",
        "oxidation_energy",
        "oxidation_free_energy",
        "ox_molecule_id",
        "ox_property_id",
        "reduction_potential",
        "oxidation_potential",
    ],
    "emmet.core.molecules.summary.MoleculeSummaryDoc": [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "property_name",
        "property_id",
        "molecule_id",
        "deprecated",
        "deprecation_reasons",
        "level_of_theory",
        "solvent",
        "lot_solvent",
        "last_updated",
        "origins",
        "warnings",
        "molecules",
        "molecule_levels_of_theory",
        "inchi",
        "inchi_key",
        "task_ids",
        "similar_molecules",
        "constituent_molecules",
        "unique_calc_types",
        "unique_task_types",
        "unique_levels_of_theory",
        "unique_solvents",
        "unique_lot_solvents",
        "thermo",
        "vibration",
        "orbitals",
        "partial_charges",
        "partial_spins",
        "bonding",
        "multipole_moments",
        "redox",
        "metal_binding",
        "has_props",
    ],
    "emmet.core.molecules.thermo.MoleculeThermoDoc": [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "property_name",
        "property_id",
        "molecule_id",
        "deprecated",
        "deprecation_reasons",
        "level_of_theory",
        "solvent",
        "lot_solvent",
        "last_updated",
        "origins",
        "warnings",
        "electronic_energy",
        "correction",
        "base_level_of_theory",
        "base_solvent",
        "base_lot_solvent",
        "correction_level_of_theory",
        "correction_solvent",
        "correction_lot_solvent",
        "combined_lot_solvent",
        "zero_point_energy",
        "rt",
        "total_enthalpy",
        "total_entropy",
        "translational_enthalpy",
        "translational_entropy",
        "rotational_enthalpy",
        "rotational_entropy",
        "vibrational_enthalpy",
        "vibrational_entropy",
        "free_energy",
    ],
    "emmet.core.molecules.vibration.VibrationDoc": [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "property_name",
        "property_id",
        "molecule_id",
        "deprecated",
        "deprecation_reasons",
        "level_of_theory",
        "solvent",
        "lot_solvent",
        "last_updated",
        "origins",
        "warnings",
        "molecule",
        "frequencies",
        "frequency_modes",
        "ir_intensities",
        "ir_activities",
        "raman_intensities",
        "raman_activities",
    ],
    "emmet.core.molecules_jcesr.MoleculesDoc": [
        "elements",
        "nelements",
        "EA",
        "IE",
        "charge",
        "pointgroup",
        "smiles",
        "task_id",
        "molecule",
        "formula_pretty",
        "svg",
    ],
    "emmet.core.oxidation_states.OxidationStateDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "possible_species",
        "possible_valences",
        "average_oxidation_states",
        "method",
    ],
    "emmet.core.phonon.PhononBSDOSDoc": [
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "identifier",
        "phonon_method",
        "phonon_bandstructure",
        "phonon_dos",
        "epsilon_static",
        "epsilon_electronic",
        "born",
        "force_constants",
        "last_updated",
        "sum_rules_breaking",
        "structure",
        "total_dft_energy",
        "volume_per_formula_unit",
        "formula_units",
        "supercell_matrix",
        "primitive_matrix",
        "code",
        "post_process_settings",
        "thermal_displacement_data",
        "calc_meta",
        "material_id",
        "task_ids",
    ],
    "emmet.core.polar.DielectricDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "total",
        "ionic",
        "electronic",
        "e_total",
        "e_ionic",
        "e_electronic",
        "n",
    ],
    "emmet.core.polar.PiezoelectricDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "total",
        "ionic",
        "electronic",
        "e_ij_max",
        "max_direction",
        "strain_for_max",
    ],
    "emmet.core.provenance.ProvenanceDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "created_at",
        "references",
        "authors",
        "remarks",
        "tags",
        "theoretical",
        "database_IDs",
        "history",
    ],
    "emmet.core.qchem.molecule.MoleculeDoc": [
        "builder_meta",
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "molecule_id",
        "molecule",
        "deprecated",
        "deprecation_reasons",
        "initial_molecules",
        "task_ids",
        "deprecated_tasks",
        "calc_types",
        "last_updated",
        "created_at",
        "origins",
        "warnings",
        "species",
        "molecules",
        "molecule_levels_of_theory",
        "inchi",
        "inchi_key",
        "task_types",
        "levels_of_theory",
        "solvents",
        "lot_solvents",
        "unique_calc_types",
        "unique_task_types",
        "unique_levels_of_theory",
        "unique_solvents",
        "unique_lot_solvents",
        "entries",
        "best_entries",
        "constituent_molecules",
        "similar_molecules",
    ],
    "emmet.core.qchem.task.TaskDocument": [
        "charge",
        "spin_multiplicity",
        "natoms",
        "elements",
        "nelements",
        "nelectrons",
        "composition",
        "composition_reduced",
        "formula_alphabetical",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "symmetry",
        "species_hash",
        "coord_hash",
        "builder_meta",
        "calc_code",
        "version",
        "dir_name",
        "task_id",
        "completed",
        "completed_at",
        "last_updated",
        "is_valid",
        "state",
        "cputime",
        "walltime",
        "calcs_reversed",
        "orig",
        "output",
        "critic2",
        "custom_smd",
        "special_run_type",
        "smiles",
        "tags",
        "warnings",
    ],
    "emmet.core.robocrys.RobocrystallogapherDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "description",
        "condensed_structure",
        "robocrys_version",
    ],
    "emmet.core.similarity.SimilarityDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "sim",
        "feature_vector",
    ],
    "emmet.core.substrates.SubstratesDoc": [
        "sub_form",
        "sub_id",
        "film_orient",
        "area",
        "energy",
        "film_id",
        "norients",
        "orient",
    ],
    "emmet.core.summary.SummaryDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "task_ids",
        "uncorrected_energy_per_atom",
        "energy_per_atom",
        "formation_energy_per_atom",
        "energy_above_hull",
        "is_stable",
        "equilibrium_reaction_energy_per_atom",
        "decomposes_to",
        "xas",
        "grain_boundaries",
        "band_gap",
        "cbm",
        "vbm",
        "efermi",
        "is_gap_direct",
        "is_metal",
        "es_source_calc_id",
        "bandstructure",
        "dos",
        "dos_energy_up",
        "dos_energy_down",
        "is_magnetic",
        "ordering",
        "total_magnetization",
        "total_magnetization_normalized_vol",
        "total_magnetization_normalized_formula_units",
        "num_magnetic_sites",
        "num_unique_magnetic_sites",
        "types_of_magnetic_species",
        "bulk_modulus",
        "shear_modulus",
        "universal_anisotropy",
        "homogeneous_poisson",
        "e_total",
        "e_ionic",
        "e_electronic",
        "n",
        "e_ij_max",
        "weighted_surface_energy_EV_PER_ANG2",
        "weighted_surface_energy",
        "weighted_work_function",
        "surface_anisotropy",
        "shape_factor",
        "has_reconstructed",
        "possible_species",
        "has_props",
        "theoretical",
        "database_IDs",
    ],
    "emmet.core.surface_properties.SurfacePropDoc": [
        "surfaces",
        "weighted_surface_energy_EV_PER_ANG2",
        "weighted_surface_energy",
        "surface_anisotropy",
        "pretty_formula",
        "shape_factor",
        "weighted_work_function",
        "has_reconstructed",
        "material_id",
        "structure",
    ],
    "emmet.core.synthesis.core.SynthesisSearchResultModel": [
        "doi",
        "paragraph_string",
        "synthesis_type",
        "reaction_string",
        "reaction",
        "target",
        "targets_formula",
        "precursors_formula",
        "targets_formula_s",
        "precursors_formula_s",
        "precursors",
        "operations",
        "search_score",
        "highlights",
    ],
    "emmet.core.tasks.TaskDoc": [
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "batch_id",
        "calc_type",
        "completed_at",
        "dir_name",
        "icsd_id",
        "input",
        "last_updated",
        "orig_inputs",
        "output",
        "run_type",
        "structure",
        "tags",
        "task_id",
        "task_type",
        "transformations",
        "vasp_objects",
        "additional_json",
        "analysis",
        "author",
        "calcs_reversed",
        "custodian",
        "entry",
        "included_objects",
        "run_stats",
        "state",
        "task_label",
    ],
    "emmet.core.thermo.ThermoDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "property_name",
        "thermo_type",
        "thermo_id",
        "uncorrected_energy_per_atom",
        "energy_per_atom",
        "energy_uncertainy_per_atom",
        "formation_energy_per_atom",
        "energy_above_hull",
        "is_stable",
        "equilibrium_reaction_energy_per_atom",
        "decomposes_to",
        "decomposition_enthalpy",
        "decomposition_enthalpy_decomposes_to",
        "energy_type",
        "entry_types",
        "entries",
    ],
    "emmet.core.vasp.material.MaterialsDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "material_id",
        "deprecated",
        "deprecation_reasons",
        "last_updated",
        "origins",
        "warnings",
        "structure",
        "initial_structures",
        "task_ids",
        "deprecated_tasks",
        "calc_types",
        "created_at",
        "task_types",
        "run_types",
        "entries",
    ],
    "emmet.core.xas.XASDoc": [
        "builder_meta",
        "nsites",
        "elements",
        "nelements",
        "composition",
        "composition_reduced",
        "formula_pretty",
        "formula_anonymous",
        "chemsys",
        "volume",
        "density",
        "density_atomic",
        "symmetry",
        "spectrum_name",
        "material_id",
        "spectrum_id",
        "last_updated",
        "warnings",
        "spectrum",
        "task_ids",
        "absorbing_element",
        "spectrum_type",
        "edge",
    ],
}


@pytest.mark.parametrize("import_str,ref_fields", ref_model_fields.items())
def test_model_field_drift(import_str, ref_fields):
    module, class_name = import_str.rsplit(".", 1)
    model_class = getattr(import_module(module), class_name)
    assert set(model_class.model_fields) == set(ref_fields)



from typing import Annotated, Any, TypeAlias, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.io.vasp.inputs import Kpoints
from typing_extensions import NotRequired, TypedDict

TypedKpointsDict = TypedDict(
    "TypedKpointsDict",
    {
        "@module": str,
        "@class": str,
        "comment": str,
        "coord_type": str,
        "generation_style": str,
        "kpoints": list[list[int]],
        "kpts_weights": list[float],
        "labels": list[str],
        "nkpoints": int,
        # "tet_connections": tuple[float, list[float]],
        "sym_weight": NotRequired[list[float]],  # tet_connections
        "tet_vertices": NotRequired[list[list[float]]],  # tet_connections
        "tet_number": int,
        "tet_weight": float,
        "usershift": list[float],
    },
)

KpointsTypeVar = TypeVar("KpointsTypeVar", Kpoints, TypedKpointsDict)


def kpoints_deserializer(kpoints: KpointsTypeVar):
    if isinstance(kpoints, dict):
        if "sym_weight" in kpoints:
            kpoints["tet_connections"] = [  # type: ignore[typeddict-unknown-key]
                (sym, vert)
                for sym, vert in zip(kpoints["sym_weight"], kpoints["tet_vertices"])
            ]
            del kpoints["sym_weight"]
            del kpoints["tet_vertices"]

        return Kpoints.from_dict(kpoints)  # type: ignore[arg-type]

    return kpoints


def kpoints_serializer(kpoints, nxt, info):
    default_serialized_object = kpoints.as_dict()

    format = info.context.get("format") if info.context else None
    if format == "arrow" and default_serialized_object:
        if default_serialized_object["tet_connections"]:
            default_serialized_object["sym_weight"] = [
                conn[0] for conn in default_serialized_object["tet_connections"]
            ]
            default_serialized_object["tet_vertices"] = [
                conn[1] for conn in default_serialized_object["tet_connections"]
            ]

        del default_serialized_object["tet_connections"]

    return default_serialized_object


KpointsType: TypeAlias = Annotated[
    KpointsTypeVar,
    BeforeValidator(kpoints_deserializer),
    WrapSerializer(kpoints_serializer, return_type=dict[str, Any]),
]



from __future__ import annotations

import copy
from hashlib import blake2b
from typing import TYPE_CHECKING, Any

import networkx as nx
from pydantic import Field
from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.core.structure import Molecule

from emmet.core.molecules import MolPropertyOrigin
from emmet.core.molecules.molecule_property import PropertyDoc
from emmet.core.mpid import MPculeID
from emmet.core.qchem.task import TaskDocument
from emmet.core.types.pymatgen_types.structure_graph_adapter import MoleculeGraphType
from emmet.core.utils import make_mol_graph

if TYPE_CHECKING:
    from typing import Any


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


metals = [
    "Li",
    "Be",
    "Na",
    "Mg",
    "Al",
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Cs",
    "Ba",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
]

BOND_METHODS = ["nbo", "critic2", "OpenBabelNN + metal_edge_extender"]


def fix_C_Li_bonds(critic: dict) -> dict:
    """
    Adjust C-Li coordinate bonding for Critic2 calculations.

    :param critic: Critic2 output dictionary

    :return:
        critic: modified Critic2 output dictionary

    """
    for key in critic["bonding"]:
        if critic["bonding"][key]["atoms"] == ["Li", "C"] or critic["bonding"][key][
            "atoms"
        ] == ["C", "Li"]:
            if (
                critic["bonding"][key]["field"] <= 0.02
                and critic["bonding"][key]["field"] > 0.012
                and critic["bonding"][key]["distance"] < 2.5
            ):
                critic["processed"]["bonds"].append(
                    [int(entry) - 1 for entry in critic["bonding"][key]["atom_ids"]]
                )
    return critic


def _bonds_hybridization(nbo: dict[str, Any], index: int):
    """
    Extract bonds from "hybridization_character" NBO output
    """

    bonds = set()
    warnings = set()

    if len(nbo["hybridization_character"]) > index:
        for bond_ind in nbo["hybridization_character"][index].get("type", list()):
            if nbo["hybridization_character"][index]["type"][bond_ind] != "BD":
                continue

            from_ind = (
                int(nbo["hybridization_character"][index]["atom 1 number"][bond_ind])
                - 1
            )
            to_ind = (
                int(nbo["hybridization_character"][index]["atom 2 number"][bond_ind])
                - 1
            )

            if (
                nbo["hybridization_character"][index]["atom 1 symbol"][bond_ind]
                in metals
            ):
                m_contrib = float(
                    nbo["hybridization_character"][index]["atom 1 polarization"][
                        bond_ind
                    ]
                )
            elif (
                nbo["hybridization_character"][index]["atom 2 symbol"][bond_ind]
                in metals
            ):
                m_contrib = float(
                    nbo["hybridization_character"][index]["atom 2 polarization"][
                        bond_ind
                    ]
                )
            else:
                m_contrib = None

            if m_contrib is None:
                bond_type = "covalent"
            elif m_contrib >= 30.0:
                bond_type = "covalent"
                warnings.add("Contains covalent bond with metal atom")
            else:
                bond_type = "electrostatic"

            bonds.add((from_ind, to_ind, bond_type))

    return bonds, warnings


def _bonds_peturbation(
    nbo: dict[str, Any],
    index: int,
    poss_coord: dict[int | None, list[int | None]],
    energy_cutoff: float,
    metal_indices: list[int],
):
    """
    Extract bonds from "perturbation_energy" NBO output
    """

    bonds = set()  # type: ignore

    # No metals, so don't need to use perturbation analysis to get bonds
    if len(metal_indices) == 0:
        return bonds

    if len(nbo["perturbation_energy"]) > index:
        for inter_ind in nbo["perturbation_energy"][index].get("donor type", list()):
            coord = False
            m_ind: int | None = None
            x_ind: int | None = None
            if (
                int(
                    nbo["perturbation_energy"][index]["acceptor atom 1 number"][
                        inter_ind
                    ]
                )
                - 1
                in metal_indices
            ):
                if (
                    nbo["perturbation_energy"][index]["donor type"][inter_ind] == "LP"
                    and nbo["perturbation_energy"][index]["acceptor type"][inter_ind]
                    == "LV"
                ):
                    coord = True
                    m_ind = (
                        int(
                            nbo["perturbation_energy"][index]["acceptor atom 1 number"][
                                inter_ind
                            ]
                        )
                        - 1
                    )
                    x_ind = (
                        int(
                            nbo["perturbation_energy"][index]["donor atom 1 number"][
                                inter_ind
                            ]
                        )
                        - 1
                    )
                elif (
                    nbo["perturbation_energy"][index]["donor type"][inter_ind] == "LP"
                    and nbo["perturbation_energy"][index]["acceptor type"][inter_ind]
                    == "RY*"
                ):
                    coord = True
                    m_ind = (
                        int(
                            nbo["perturbation_energy"][index]["acceptor atom 1 number"][
                                inter_ind
                            ]
                        )
                        - 1
                    )
                    x_ind = (
                        int(
                            nbo["perturbation_energy"][index]["donor atom 1 number"][
                                inter_ind
                            ]
                        )
                        - 1
                    )
            elif (
                nbo["perturbation_energy"][index]["donor atom 1 number"][inter_ind] - 1
                in metal_indices
            ):
                if (
                    nbo["perturbation_energy"][index]["donor type"][inter_ind] == "LP"
                    and nbo["perturbation_energy"][index]["acceptor type"][inter_ind]
                    == "LV"
                ):
                    coord = True
                    m_ind = (
                        int(
                            nbo["perturbation_energy"][index]["donor atom 1 number"][
                                inter_ind
                            ]
                        )
                        - 1
                    )
                    x_ind = (
                        int(
                            nbo["perturbation_energy"][index]["acceptor atom 1 number"][
                                inter_ind
                            ]
                        )
                        - 1
                    )

            if not coord:
                continue
            elif x_ind not in poss_coord[m_ind]:
                continue

            energy = float(
                nbo["perturbation_energy"][index]["perturbation energy"][inter_ind]
            )
            if energy >= energy_cutoff:
                bonds.add((x_ind, m_ind, "electrostatic"))
    return bonds


def nbo_molecule_graph(mol: Molecule, nbo: dict[str, Any]):
    """
    Construct a molecule graph from NBO data.

    :param mol: molecule to be analyzed
    :param nbo: Output from NBO7
    :return:
    """

    mg = MoleculeGraph.from_empty_graph(mol)

    alpha_bonds, warnings = _bonds_hybridization(nbo, 1)
    beta_bonds, new_warnings = _bonds_hybridization(nbo, 3)
    warnings = warnings.union(new_warnings)

    distance_cutoff = 3.0
    energy_cutoff = 3.0
    metal_indices = [i for i, e in enumerate(mol.species) if str(e) in metals]

    poss_coord: dict[int | None, list[int | None]] = dict()
    dist_mat = mol.distance_matrix
    for i in metal_indices:
        poss_coord[i] = list()
        row = dist_mat[i]
        for j, val in enumerate(row):
            if i != j and val < distance_cutoff:
                poss_coord[i].append(j)

    new_alpha_bonds = _bonds_peturbation(
        nbo, 0, poss_coord, energy_cutoff, metal_indices
    )
    alpha_bonds = alpha_bonds.union(new_alpha_bonds)

    if mol.spin_multiplicity != 1:
        new_beta_bonds = _bonds_peturbation(
            nbo, 1, poss_coord, energy_cutoff, metal_indices
        )
        beta_bonds = beta_bonds.union(new_beta_bonds)

    sorted_alpha = set([tuple(sorted([a[0], a[1]])) for a in alpha_bonds])
    sorted_beta = set([tuple(sorted([b[0], b[1]])) for b in beta_bonds])

    if sorted_alpha != sorted_beta:
        warnings.add("Difference in bonding between alpha and beta electrons")

    for bond in alpha_bonds.union(beta_bonds):
        if (bond[0], bond[1]) not in mg.graph.edges() and (
            bond[1],
            bond[0],
        ) not in mg.graph.edges():
            if bond[0] < bond[1]:
                mg.add_edge(bond[0], bond[1], edge_properties={"type": bond[2]})
            else:
                mg.add_edge(bond[1], bond[0], edge_properties={"type": bond[2]})

    mg_copy = copy.deepcopy(mg)
    mg_copy.remove_nodes(metal_indices)

    try:
        if not nx.is_connected(mg_copy.graph.to_undirected()):
            warnings.add("Metal-centered complex")
    except nx.exception.NetworkXPointlessConcept:
        if len(mg.molecule) == 1:
            warnings.add("Single-atom; no bonds")

    return (mg, list(warnings))


class MoleculeBondingDoc(PropertyDoc):
    """Representation of molecular bonding."""

    property_name: str = "bonding"

    molecule_graph: MoleculeGraphType = Field(..., description="Molecule graph")

    method: str = Field(..., description="Method used to compute molecule graph")

    bond_types: dict[str, list[float]] = Field(
        dict(),
        description="Dictionary of bond types to their length, e.g. C-O to "
        "a list of the lengths of C-O bonds in Angstrom.",
    )

    bonds: list[tuple[int, int]] = Field(
        [],
        description="List of bonds in the form (a, b), where a and b are 0-indexed atom indices",
    )

    bonds_nometal: list[tuple[int, int]] = Field(
        [],
        description="List of bonds in the form (a, b), where a and b are 0-indexed atom indices, "
        "with all metal ions removed",
    )

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        preferred_methods: list[str],
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Determine bonding from a task document

        Method preferences are as follows:
        - NBO7
        - OpenBabelNN + metal_edge_extender in pymatgen
        - Critic2 (really OpenBabelNN + metal_edge_extender + Critic2)

        :param task: task document from which bonding properties can be extracted
        :param molecule_id: MPculeID
        :param preferred_methods: list of methods; by default, NBO7, Critic2, and the combination
            of OpenBabelNN and metal_edge_extender in pymatgen, in that order
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        mg_made = False
        method = None
        warnings = list()

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        for m in preferred_methods:
            if mg_made:
                break

            if (
                m == "nbo"
                and task.output.nbo is not None
                and (
                    task.orig["rem"].get("run_nbo6", False)
                    or task.orig["rem"].get("nbo_external", False)
                )
            ):
                method = "nbo"
                mg, warnings = nbo_molecule_graph(mol, task.output.nbo)
                mg_made = True

            elif m == "critic2" and task.critic2 is not None:
                method = "critic2"
                critic = fix_C_Li_bonds(task.critic2)
                critic_bonds = critic["processed"]["bonds"]
                mg = make_mol_graph(mol, critic_bonds=critic_bonds)
                mg_made = True

            else:
                method = "OpenBabelNN + metal_edge_extender"
                mg = make_mol_graph(mol)
                mg_made = True

        bonds = list()
        for bond in mg.graph.edges():
            bonds.append(sorted([bond[0], bond[1]]))

        # Calculate bond lengths
        bond_types = dict()
        for u, v in mg.graph.edges():
            species_u = str(mg.molecule.species[u])
            species_v = str(mg.molecule.species[v])
            if species_u < species_v:
                species = f"{species_u}-{species_v}"
            else:
                species = f"{species_v}-{species_u}"
            dist = mg.molecule.get_distance(u, v)
            if species not in bond_types:
                bond_types[species] = [dist]
            else:
                bond_types[species].append(dist)

        m_inds = [e for e in range(len(mol)) if str(mol.species[e]) in metals]

        bonds_nometal = list()
        for bond in bonds:
            if not any([m in bond for m in m_inds]):
                bonds_nometal.append(bond)

        id_string = f"bonding-{molecule_id}-{task.task_id}-{task.lot_solvent}-{method}"
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        return super().from_molecule(
            meta_molecule=mol,
            property_id=property_id,
            molecule_id=molecule_id,
            level_of_theory=task.level_of_theory,
            solvent=task.solvent,
            lot_solvent=task.lot_solvent,
            method=method,
            warnings=warnings,
            molecule_graph=mg,
            bond_types=bond_types,
            bonds=bonds,
            bonds_nometal=bonds_nometal,
            origins=[MolPropertyOrigin(name="bonding", task_id=task.task_id)],
            deprecated=deprecated,
            **kwargs,
        )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.analysis.graphs import MoleculeGraph, StructureGraph
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedMoleculeDict,
    TypedStructureDict,
    pop_empty_structure_keys,
)

GraphDescriptors = list[str, str]  # type: ignore[type-arg]


class TypedNodeDict(TypedDict):
    id: int


class TypedAdjacencyDict(TypedDict):
    to_jimage: list[int, int, int]  # type: ignore[type-arg]
    weight: float
    id: int
    key: int


class TypedGraphDict(TypedDict):
    directed: bool
    multigraph: bool
    graph: list[GraphDescriptors, GraphDescriptors, GraphDescriptors]  # type: ignore[type-arg]
    nodes: list[TypedNodeDict]
    adjacency: list[list[TypedAdjacencyDict]]


TypedStructureGraphDict = TypedDict(
    "TypedStructureGraphDict",
    {
        "@module": str,
        "@class": str,
        "structure": TypedStructureDict,
        "graphs": TypedGraphDict,
    },
)

TypedMoleculeGraphDict = TypedDict(
    "TypedMoleculeGraphDict",
    {
        "@module": str,
        "@class": str,
        "molecule": TypedMoleculeDict,
        "graphs": TypedGraphDict,
    },
)

StructureGraphTypeVar = TypeVar(
    "StructureGraphTypeVar", StructureGraph, TypedStructureGraphDict
)


MoleculeGraphTypeVar = TypeVar(
    "MoleculeGraphTypeVar", MoleculeGraph, TypedMoleculeGraphDict
)


def pop_empty_graph_keys(graph: StructureGraphTypeVar | MoleculeGraphTypeVar):
    if isinstance(graph, dict):
        target_cls: type[StructureGraph | MoleculeGraph]
        match graph["@class"]:
            case "StructureGraph":
                target_cls = StructureGraph
                graph["structure"] = pop_empty_structure_keys(graph["structure"])  # type: ignore[typeddict-unknown-key, typeddict-item]
            case "MoleculeGraph":
                target_cls = MoleculeGraph
                graph["molecule"] = pop_empty_structure_keys(graph["molecule"])  # type: ignore[typeddict-unknown-key, typeddict-item]

        return target_cls.from_dict(graph)  # type: ignore[arg-type]

    return graph


StructureGraphType = Annotated[
    StructureGraphTypeVar,
    BeforeValidator(pop_empty_graph_keys),
    WrapSerializer(
        lambda x, nxt, info: x.as_dict(), return_type=TypedStructureGraphDict
    ),
]

MoleculeGraphType = Annotated[
    MoleculeGraphTypeVar,
    BeforeValidator(pop_empty_graph_keys),
    WrapSerializer(
        lambda x, nxt, info: x.as_dict(), return_type=TypedMoleculeGraphDict
    ),
]



"""Core definitions of a QChem calculation document."""

from __future__ import annotations

import logging
import re
import warnings
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import Any

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, field_validator
from pymatgen.core.structure import Molecule
from pymatgen.io.qchem.inputs import QCInput
from pymatgen.io.qchem.outputs import QCOutput

from emmet.core.qchem.calc_types import CalcType, LevelOfTheory, TaskType
from emmet.core.qchem.calc_types.calc_types import BASIS_SETS, FUNCTIONALS
from emmet.core.qchem.task import QChemStatus
from emmet.core.utils import arrow_incompatible

# from emmet.core.qchem.calc_types.em_utils import (
#     level_of_theory,
#     task_type,
#     calc_type,
# )


functional_synonyms = {
    "b97mv": "b97m-v",
    "b97mrv": "b97m-rv",
    "wb97xd": "wb97x-d",
    "wb97xd3": "wb97x-d3",
    "wb97xv": "wb97x-v",
    "wb97mv": "wb97m-v",
}

smd_synonyms = {
    "DIELECTRIC=7,230;N=1,410;ALPHA=0,000;BETA=0,859;GAMMA=36,830;PHI=0,000;PSI=0,000": "diglyme",
    "DIELECTRIC=18,500;N=1,415;ALPHA=0,000;BETA=0,735;GAMMA=20,200;PHI=0,000;PSI=0,000": "3:7 EC:EMC",
}

__author__ = "Rishabh D. Guha <rdguha@lbl.gov>"
logger = logging.getLogger(__name__)

# class QChemObject(ValueEnum):
# Not sure but can we have something like GRAD and HESS
# as QChem data objects


@arrow_incompatible
class CalculationInput(BaseModel):
    """
    Document defining QChem calculation inputs.
    """

    initial_molecule: Molecule | None = Field(
        None, description="input molecule geometry before the QChem calculation"
    )

    # parameters: dict[str, Any] = Field(
    #     None, description = "Parameters from a previous QChem calculation."
    # )

    charge: int | None = Field(None, description="The charge of the input molecule")

    rem: dict[str, Any] | None = Field(
        None,
        description="The rem dict of the input file which has all the input parameters",
    )

    job_type: str | None = Field(
        None, description="The type of QChem calculation being performed"
    )

    opt: dict[str, Any] | None = Field(
        None,
        description="A dictionary of opt section. For instance atom constraints and fixed atoms. Go to QCInput definition for more details.",
    )

    pcm: dict[str, Any] | None = Field(
        None, description="A dictionary for the PCM solvent details if used"
    )

    solvent: dict[str, Any] | None = Field(
        None,
        description="The solvent parameters used if the PCM solvent model has been employed",
    )

    smx: dict[str, Any] | None = Field(
        None,
        description="A dictionary for the solvent parameters if the SMD solvent method has been employed",
    )

    vdw_mode: str | None = Field(
        None,
        description="Either atomic or sequential. Used when custon van der Waals radii are used to construct pcm cavity",
    )

    van_der_waals: dict[str, Any] | None = Field(
        None, description="The dictionary of the custom van der Waals radii if used"
    )

    scan_variables: dict[str, Any] | None = Field(
        None,
        description="The dictionary of scan variables for torsions or bond stretches",
    )

    tags: dict[str, Any] | str | None = Field(
        None, description="Any tags associated with the QChem calculation."
    )

    @classmethod
    def from_qcinput(cls, qcinput: QCInput) -> "CalculationInput":
        """
        Create a QChem input document from a QCInout object.

        Parameters
        ----------
        qcinput
            A QCInput object.

        Returns
        --------
        CalculationInput
            The input document.
        """

        # QCInput.molecule is either a Molecule, list of Molecule, or a str-literal
        # Catch latter two cases here and error out
        if not isinstance(qcinput.molecule, Molecule):
            raise ValueError("QCInput contains multiple molecules.")

        return cls(  # type: ignore[call-arg]
            initial_molecule=qcinput.molecule,
            charge=(
                int(qcinput.molecule.as_dict()["charge"])
                if qcinput.molecule.as_dict()
                else None
            ),
            rem=qcinput.rem,
            job_type=qcinput.rem.get("job_type", None),
            opt=qcinput.opt,
            pcm=qcinput.pcm,
            solvent=qcinput.solvent,
            smx=qcinput.smx,
            vdw_mode=qcinput.vdw_mode,
            scan_variables=qcinput.scan,
            van_der_waals=qcinput.van_der_waals,
        )


@arrow_incompatible
class CalculationOutput(BaseModel):
    """Document defining QChem calculation outputs."""

    optimized_molecule: Molecule | None = Field(
        None,
        description="optimized geometry of the molecule after calculation in case of opt, optimization or ts",
    )

    mulliken: list | dict[str, Any] | None = Field(
        None, description="Calculate Mulliken charges on the atoms"
    )

    esp: list | dict[str, Any] | None = Field(
        None,
        description="Calculated charges on the atoms if esp calculation has been performed",
    )

    resp: list | dict[str, Any] | None = Field(
        None,
        description="Calculated charges on the atoms if resp calculation has been performed",
    )

    nbo_data: dict[str, Any] | None = Field(
        None, description="NBO data if analysis has been performed."
    )

    frequencies: dict[str, Any] | list | None = Field(
        None,
        description="Calculated frequency modes if the job type is freq or frequency",
    )

    frequency_modes: list | str | None = Field(
        None, description="The list of calculated frequency mode vectors"
    )

    final_energy: str | float | None = Field(
        None,
        description="The final energy of the molecule after the calculation is complete",
    )

    enthalpy: str | float | None = Field(
        None,
        description="The total enthalpy correction if a frequency calculation has been performed",
    )

    entropy: str | float | None = Field(
        None,
        description="The total entropy of the system if a frequency calculation has been performed",
    )

    scan_energies: dict[str, Any] | None = Field(
        None,
        description="A dictionary of the scan energies with their respective parameters",
    )

    scan_geometries: list | dict[str, Any] | None = Field(
        None, description="optimized geometry of the molecules after the geometric scan"
    )

    scan_molecules: list | dict[str, Any] | Molecule | None = Field(
        None,
        description="The constructed pymatgen molecules from the optimized scan geometries",
    )

    pcm_gradients: dict[str, Any] | np.ndarray | list | None = Field(
        None,
        description="The parsed total gradients after adding the PCM contributions.",
    )

    @field_validator("pcm_gradients", mode="before")
    @classmethod
    def validate_pcm_gradients(cls, v):
        if v is not None and not isinstance(v, (np.ndarray, dict, list)):
            raise ValueError(
                "pcm_gradients must be a numpy array, a dict or a list or None."
            )
        return v

    cds_gradients: dict[str, Any] | np.ndarray | list | None = Field(
        None, description="The parsed CDS gradients."
    )

    @field_validator("cds_gradients", mode="before")
    @classmethod
    def validate_cds_gradients(cls, v):
        if v is not None and not isinstance(v, (np.ndarray, dict, list)):
            raise ValueError(
                "cds_gradients must be a numpy array, a dict or a list or None."
            )
        return v

    dipoles: dict[str, Any] | None = Field(
        None, description="The associated dipoles for Mulliken/RESP/ESP charges"
    )

    gap_info: dict[str, Any] | None = Field(
        None, description="The Kohn-Sham HOMO-LUMO gaps and associated Eigenvalues"
    )

    @classmethod
    def from_qcoutput(cls, qcoutput: QCOutput) -> "CalculationOutput":
        """
        Create a QChem output document from a QCOutput object.

        Parameters
        ----------
        qcoutput
            A QCOutput object.

        Returns
        --------
        CalculationOutput
            The output document.
        """
        optimized_molecule = qcoutput.data.get("molecule_from_optimized_geometry", {})
        optimized_molecule = optimized_molecule if optimized_molecule else None
        return cls(
            optimized_molecule=optimized_molecule,
            mulliken=qcoutput.data.get(["Mulliken"][-1], []),
            esp=qcoutput.data.get(["ESP"][-1], []),
            resp=qcoutput.data.get(["RESP"][-1], []),
            nbo_data=qcoutput.data.get("nbo_data", {}),
            frequencies=qcoutput.data.get("frequencies", {}),
            frequency_modes=qcoutput.data.get("frequency_mode_vectors", []),
            final_energy=qcoutput.data.get("final_energy", None),
            enthalpy=qcoutput.data.get("total_enthalpy", None),
            entropy=qcoutput.data.get("total_entropy", None),
            scan_energies=qcoutput.data.get("scan_energies", {}),
            scan_geometries=qcoutput.data.get("optimized_geometries", {}),
            scan_molecules=qcoutput.data.get("molecules_from_optimized_geometries", {}),
            pcm_gradients=qcoutput.data.get(["pcm_gradients"][0], None),
            cds_gradients=qcoutput.data.get(["CDS_gradients"][0], None),
            dipoles=qcoutput.data.get("dipoles", None),
            gap_info=qcoutput.data.get("gap_info", None),
        )

    model_config = ConfigDict(arbitrary_types_allowed=True)
    # TODO What can be done for the trajectories, also how will walltime and cputime be reconciled


@arrow_incompatible
class Calculation(BaseModel):
    """Full QChem calculation inputs and outputs."""

    dir_name: str = Field(description="The directory for this QChem calculation")
    qchem_version: str | None = Field(
        None, description="QChem version used to perform the current calculation"
    )
    has_qchem_completed: QChemStatus | bool | None = Field(
        None, description="Whether QChem calculated the calculation successfully"
    )
    input: CalculationInput | None = Field(
        None, description="QChem input settings for the calculation"
    )
    output: CalculationOutput | None = Field(
        None, description="The QChem calculation output document"
    )
    completed_at: str | None = Field(
        None, description="Timestamp for when the calculation was completed"
    )
    task_name: str | None = Field(
        None,
        description="Name of task given by custodian (e.g. opt1, opt2, freq1, freq2)",
    )
    output_file_paths: dict[str, str | Path | dict[str, Path]] | None = Field(
        None,
        description="Paths (relative to dir_name) of the QChem output files associated with this calculation",
    )
    level_of_theory: LevelOfTheory | str = Field(
        None,
        description="Levels of theory used for the QChem calculation: For instance, B97-D/6-31g*",
    )
    solvation_lot_info: str | None = Field(
        None,
        description="A condensed string representation of the comboned LOT and Solvent info",
    )
    task_type: TaskType | None = Field(
        None,
        description="Calculation task type like Single Point, Geometry Optimization. Frequency...",
    )
    calc_type: CalcType | str | None = Field(
        None,
        description="Combination dict of LOT + TaskType: B97-D/6-31g*/VACUUM Geometry Optimization",
    )

    @classmethod
    def from_qchem_files(
        cls,
        dir_name: Path | str,
        task_name: str,
        qcinput_file: Path | str,
        qcoutput_file: Path | str,
        validate_lot: bool = True,
        store_energy_trajectory: bool = False,
        qcinput_kwargs: dict | None = None,
        qcoutput_kwargs: dict | None = None,
    ) -> "Calculation":
        """
        Create a QChem calculation document from a directory and file paths.

        Parameters
        ----------
        dir_name
            The directory containing the QChem calculation outputs.
        task_name
            The task name.
        qcinput_file
            Path to the .in/qin file, relative to dir_name.
        qcoutput_file
            Path to the .out/.qout file, relative to dir_name.
        store_energy_trajectory
            Whether to store the energy trajectory during a QChem calculation #TODO: Revisit this- False for now.
        qcinput_kwargs
            Additional keyword arguments that will be passed to the qcinput file
        qcoutput_kwargs
            Additional keyword arguments that will be passed to the qcoutput file

        Returns
        -------
        Calculation
            A QChem calculation document.
        """

        dir_name = Path(dir_name)
        qcinput_file = dir_name / qcinput_file
        qcoutput_file = dir_name / qcoutput_file

        output_file_paths = _find_qchem_files(dir_name)

        qcinput_kwargs = qcinput_kwargs if qcinput_kwargs else {}
        qcinput = QCInput.from_file(qcinput_file, **qcinput_kwargs)

        qcoutput_kwargs = qcoutput_kwargs if qcoutput_kwargs else {}
        qcoutput = QCOutput(str(qcoutput_file), **qcoutput_kwargs)

        completed_at = str(datetime.fromtimestamp(qcoutput_file.stat().st_mtime))

        input_doc = CalculationInput.from_qcinput(qcinput)
        output_doc = CalculationOutput.from_qcoutput(qcoutput)

        has_qchem_completed = (
            QChemStatus.SUCCESS
            if qcoutput.data.get("completion", [])
            else QChemStatus.FAILED
        )

        if store_energy_trajectory:
            warnings.warn("Still have to figure the energy trajectory")

        return cls(
            dir_name=str(dir_name),
            task_name=task_name,
            qchem_version=qcoutput.data["version"],
            has_qchem_completed=has_qchem_completed,
            completed_at=completed_at,
            input=input_doc,
            output=output_doc,
            output_file_paths={
                k.lower(): (
                    Path(v)
                    if isinstance(v, str)
                    else {k2: Path(v2) for k2, v2 in v.items()}
                )
                for k, v in output_file_paths.items()
            },
            level_of_theory=level_of_theory(input_doc, validate_lot=validate_lot),
            solvation_lot_info=lot_solvent_string(input_doc, validate_lot=validate_lot),
            task_type=task_type(input_doc),
            calc_type=calc_type(input_doc, validate_lot=validate_lot),
        )


def _find_qchem_files(
    path: str | Path,
) -> dict[str, Any]:
    """
    Find QChem files in a directory.

    Only the mol.qout file (or alternatively files
    with the task name as an extension, e.g., mol.qout.opt_0.gz, mol.qout.freq_1.gz, or something like this...)
    will be returned.

    Parameters
    ----------
    path
        Path to a directory to search.

    Returns
    -------
    dict[str, Any]
        The filenames of the calculation outputs for each QChem task, given as a ordered dictionary of::

            {
                task_name:{
                    "qchem_out_file": qcrun_filename,
                },
                ...
            }
    If there is only 1 qout file task_name will be "standard" otherwise it will be the extension name like "opt_0"
    """
    path = Path(path)
    task_files = OrderedDict()

    in_file_pattern = re.compile(r"^(?P<in_task_name>mol\.(qin|in)(?:\..+)?)(\.gz)?$")

    for file in path.iterdir():
        if file.is_file():
            in_match = in_file_pattern.match(file.name)

            # This block is for generalizing outputs coming from both atomate and manual qchem calculations
            if in_match:
                in_task_name = re.sub(
                    r"(\.gz|gz)$",
                    "",
                    in_match.group("in_task_name").replace("mol.qin.", ""),
                )
                in_task_name = in_task_name or "mol.qin"
                if in_task_name == "orig":
                    task_files[in_task_name] = {"orig_input_file": file.name}
                elif in_task_name == "last":
                    continue
                elif in_task_name == "mol.qin" or in_task_name == "mol.in":
                    if in_task_name == "mol.qin":
                        out_file = (
                            path / "mol.qout.gz"
                            if (path / "mol.qout.gz").exists()
                            else path / "mol.qout"
                        )
                    else:
                        out_file = (
                            path / "mol.out.gz"
                            if (path / "mol.out.gz").exists()
                            else path / "mol.out"
                        )
                    task_files["standard"] = {
                        "qcinput_file": file.name,
                        "qcoutput_file": out_file.name,
                    }
                # This block will exist only if calcs were run through atomate
                else:
                    try:
                        task_files[in_task_name] = {
                            "qcinput_file": file.name,
                            "qcoutput_file": Path(
                                "mol.qout." + in_task_name + ".gz"
                            ).name,
                        }
                    except FileNotFoundError:
                        task_files[in_task_name] = {
                            "qcinput_file": file.name,
                            "qcoutput_file": "No qout files exist for this in file",
                        }

    return task_files


def level_of_theory(
    parameters: CalculationInput, validate_lot: bool = True
) -> LevelOfTheory:
    """

    Returns the level of theory for a calculation,
    based on the input parameters given to Q-Chem

    Args:
        parameters: dict of Q-Chem input parameters

    """

    rem_pars = parameters.rem or {}
    funct_raw = rem_pars.get("method")
    basis_raw = rem_pars.get("basis")

    if funct_raw is None or basis_raw is None:
        raise ValueError(
            'Method and basis must be included in "rem" section ' "of parameters!"
        )

    disp_corr = rem_pars.get("dft_d")

    if disp_corr is None:
        funct_lower = funct_raw.lower()
        funct_lower = functional_synonyms.get(funct_lower, funct_lower)
    else:
        # Replace Q-Chem terms for D3 tails with more common expressions
        disp_corr = disp_corr.replace("_bj", "(bj)").replace("_zero", "(0)")
        funct_lower = f"{funct_raw}-{disp_corr}"

    basis_lower = basis_raw.lower()

    solvent_method = rem_pars.get("solvent_method", "").lower()

    if solvent_method == "":
        solvation = "VACUUM"
    elif solvent_method in ["pcm", "cosmo"]:
        solvation = "PCM"
    # TODO: Add this once added into pymatgen and atomate
    # elif solvent_method == "isosvp":
    #     if parameters.get("svp", {}).get("idefesr", 0):
    #         solvation = "CMIRS"
    #     else:
    #         solvation = "ISOSVP"
    elif solvent_method == "smd":
        solvation = "SMD"
    else:
        raise ValueError(f"Unexpected implicit solvent method {solvent_method}!")

    if validate_lot:
        # --> TODO: replace with enums
        functional = [f for f in FUNCTIONALS if f.lower() == funct_lower]
        if not functional:
            raise ValueError(f"Unexpected functional {funct_lower}!")

        functional = functional[0]

        basis = [b for b in BASIS_SETS if b.lower() == basis_lower]
        if not basis:
            raise ValueError(f"Unexpected basis set {basis_lower}!")

        basis = basis[0]
        # <--

        lot = f"{functional}/{basis}/{solvation}"

        return LevelOfTheory(lot)
    else:
        warnings.warn(
            "User has turned the validate flag off."
            "This can have downstream effects if the chosen functional and basis "
            "is not in the available sets of MP employed functionals and the user"
            "wants to include the TaskDoc in the MP infrastructure."
            "Users should ignore this warning if their objective is just to create TaskDocs",
            UserWarning,
            stacklevel=2,
        )
        functional = funct_lower
        basis = basis_lower
        lot = f"{functional}/{basis}/{solvation}"

        return lot


def solvent(
    parameters: CalculationInput,
    validate_lot: bool = True,
    custom_smd: str | None = None,
) -> str:
    """
    Returns the solvent used for this calculation.

    Args:
        parameters: dict of Q-Chem input parameters
        custom_smd: (Optional) string representing SMD parameters for a
        non-standard solvent
    """
    lot = level_of_theory(parameters, validate_lot=validate_lot)
    if validate_lot:
        solvation = lot.value.split("/")[-1]
    else:
        solvation = lot.split("/")[-1]

    if solvation == "PCM":
        # dielectric = float(parameters.get("solvent", {}).get("dielectric", 78.39))
        # dielectric = float(parameters.get("solvent", {}))
        # dielectric = getattr(parameters, "solvent", None)
        # dielectric_string = f"{dielectric.get('dielectric', '0.0'):.2f}".replace(".", ",")
        dielectric_string = getattr(parameters, "solvent", None)
        return f"DIELECTRIC= {dielectric_string}"
    # TODO: Add this once added into pymatgen and atomate
    # elif solvation == "ISOSVP":
    #     dielectric = float(parameters.get("svp", {}).get("dielst", 78.39))
    #     rho = float(parameters.get("svp", {}).get("rhoiso", 0.001))
    #     return f"DIELECTRIC={round(dielectric, 2)},RHO={round(rho, 4)}"
    # elif solvation == "CMIRS":
    #     dielectric = float(parameters.get("svp", {}).get("dielst", 78.39))
    #     rho = float(parameters.get("svp", {}).get("rhoiso", 0.001))
    #     a = parameters.get("pcm_nonels", {}).get("a")
    #     b = parameters.get("pcm_nonels", {}).get("b")
    #     c = parameters.get("pcm_nonels", {}).get("c")
    #     d = parameters.get("pcm_nonels", {}).get("d")
    #     solvrho = parameters.get("pcm_nonels", {}).get("solvrho")
    #     gamma = parameters.get("pcm_nonels", {}).get("gamma")
    #
    #     string = f"DIELECTRIC={round(dielectric, 2)},RHO={round(rho, 4)}"
    #     for name, (piece, digits) in {"A": (a, 6), "B": (b, 6), "C": (c, 1), "D": (d, 3),
    #                                   "SOLVRHO": (solvrho, 2), "GAMMA": (gamma, 1)}.items():
    #         if piece is None:
    #             piecestring = "NONE"
    #         else:
    #             piecestring = f"{name}={round(float(piece), digits)}"
    #         string += "," + piecestring
    #     return string
    elif solvation == "SMD":
        solvent = (parameters.smx or {}).get("solvent", "water")
        if solvent == "other":
            if custom_smd is None:
                raise ValueError(
                    "SMD calculation with solvent=other requires custom_smd!"
                )

            names = ["DIELECTRIC", "N", "ALPHA", "BETA", "GAMMA", "PHI", "PSI"]
            numbers = [float(x) for x in custom_smd.split(",")]

            string = ""
            for name, number in zip(names, numbers):
                string += f"{name}={number:.3f};"
            return string.rstrip(",").rstrip(";").replace(".", ",")
        else:
            return f"SOLVENT={solvent.upper()}"
    else:
        return "NONE"


def lot_solvent_string(
    parameters: CalculationInput,
    validate_lot: bool = True,
    custom_smd: str | None = None,
) -> str:
    """
    Returns a string representation of the level of theory and solvent used for this calculation.

    Args:
        parameters: dict of Q-Chem input parameters
        custom_smd: (Optional) string representing SMD parameters for a
        non-standard solvent
    """
    if validate_lot:
        lot = level_of_theory(parameters, validate_lot=validate_lot).value
    else:
        lot = level_of_theory(parameters, validate_lot=validate_lot)
    solv = solvent(parameters, custom_smd=custom_smd, validate_lot=validate_lot)
    return f"{lot}({solv})"


def task_type(
    parameters: CalculationInput, special_run_type: str | None = None
) -> TaskType:
    if special_run_type == "frequency_flattener":
        return TaskType("Frequency Flattening Geometry Optimization")
    elif special_run_type == "ts_frequency_flattener":
        return TaskType("Frequency Flattening Transition State Geometry Optimization")

    if parameters.job_type == "sp":
        return TaskType("Single Point")
    elif parameters.job_type == "force":
        return TaskType("Force")
    elif parameters.job_type == "opt":
        return TaskType("Geometry Optimization")
    elif parameters.job_type == "ts":
        return TaskType("Transition State Geometry Optimization")
    elif parameters.job_type == "freq":
        return TaskType("Frequency Analysis")

    return TaskType("Unknown")


def calc_type(
    parameters: CalculationInput,
    validate_lot: bool = True,
    special_run_type: str | None = None,
) -> CalcType:
    """
    Determines the calc type

    Args:
        parameters: CalculationInput parameters
    """
    tt = task_type(parameters, special_run_type=special_run_type).value
    if validate_lot:
        rt = level_of_theory(parameters, validate_lot=validate_lot).value
        return CalcType(f"{rt} {tt}")
    else:
        rt = level_of_theory(parameters, validate_lot=validate_lot)
        return str(f"{rt} {tt}")



import datetime
import json
from pathlib import Path

import numpy as np
import pytest
from bson.objectid import ObjectId
from monty.json import MSONable
from monty.serialization import loadfn

from emmet.core.tasks import TaskDoc
from emmet.core.utils import (
    convert_datetime,
    dynamic_import,
    get_flat_models_from_model,
    get_hash_blocked,
    jsanitize,
    utcnow,
)


def test_dt():

    ref_dt = utcnow()

    for test_dt in (
        jsanitize(ref_dt),
        {"$date": jsanitize(ref_dt)},
    ):
        assert convert_datetime(test_dt) == ref_dt


def test_jsanitize():
    """
    Tests emmet Jsanitize which converts MSONable classes into dicts
    """
    # clean_json should have no effect on None types.
    d = {"hello": 1, "world": None}
    clean = jsanitize(d)
    assert clean["world"] is None
    assert json.loads(json.dumps(d)) == json.loads(json.dumps(clean))

    d = {"hello": GoodMSONClass(1, 2, 3)}
    with pytest.raises(TypeError):
        json.dumps(d)

    clean = jsanitize(d)
    assert isinstance(clean["hello"], dict)
    clean_strict = jsanitize(d, strict=True)
    assert clean_strict["hello"]["a"] == 1
    assert clean_strict["hello"]["b"] == 2

    d = {"dt": datetime.datetime.now()}
    clean = jsanitize(d)
    assert isinstance(clean["dt"], str)
    clean = jsanitize(d, allow_bson=True)
    assert isinstance(clean["dt"], datetime.datetime)

    d = {
        "a": ["b", np.array([1, 2, 3])],
        "b": ObjectId.from_datetime(datetime.datetime.now()),
    }
    clean = jsanitize(d)
    assert clean["a"] == ["b", [1, 2, 3]]
    assert isinstance(clean["b"], str)

    rnd_bin = bytes(np.random.rand(10))
    d = {"a": bytes(rnd_bin)}
    clean = jsanitize(d, allow_bson=True)
    assert clean["a"] == bytes(rnd_bin)
    assert isinstance(clean["a"], bytes)


class GoodMSONClass(MSONable):
    def __init__(self, a, b, c, d=1, **kwargs):
        self.a = a
        self.b = b
        self._c = c
        self._d = d
        self.kwargs = kwargs

    def __eq__(self, other):
        return (
            self.a == other.a
            and self.b == other.b
            and self._c == other._c
            and self._d == other._d
            and self.kwargs == other.kwargs
        )


def test_blocked_hash(tmp_dir):
    import blake3

    file_text = (
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, "
        "sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
    ).encode()

    with open("test_md5.gz", "wb") as f:
        f.write(file_text)

    assert get_hash_blocked("test_md5.gz") == blake3.blake3(file_text).hexdigest()


def test_model_flatten():

    sub_models = get_flat_models_from_model(TaskDoc)
    assert {model.__name__ for model in sub_models} == {
        "AnalysisDoc",
        "Calculation",
        "CalculationInput",
        "CalculationOutput",
        "CustodianDoc",
        "ElectronPhononDisplacedStructures",
        "ElectronicStep",
        "FrequencyDependentDielectric",
        "IonicStep",
        "OutputDoc",
        "PotcarSpec",
        "RunStatistics",
        "SymmetryData",
        "TaskDoc",
    }


def test_import():

    assert dynamic_import("emmet.core.tasks.TaskDoc") == TaskDoc


def test_data_archive(tmp_dir):

    from monty.io import zopen
    from emmet.core.testing_utils import DataArchive

    def listify_file(file_name):
        with zopen(file_name, "rt") as f:
            return list(f.read())

    test_str = "the quick brown fox jumped over the lazy dog"
    test_files = {"test_file", "test_gzipped.gz"}
    for p in test_files:
        with zopen(p, "wt") as f:
            f.write(test_str)

    archive = DataArchive.from_directory(".")

    archive_path = Path("test.json.gz")
    archive.to_json(archive_path)
    assert archive_path.is_file()
    loaded = loadfn(archive_path)
    assert set(loaded) == test_files
    assert loaded["test_file"]["0"] == test_str.encode().hex()

    with DataArchive.extract(archive_path) as temp_dir:
        assert {p.name for p in temp_dir.glob("*")} == test_files

    for fname in test_files:
        assert DataArchive.extract_obj(archive_path, fname, listify_file) == list(
            test_str
        )



from typing import Annotated, TypeVar

from pydantic.functional_serializers import WrapSerializer
from pydantic.functional_validators import BeforeValidator
from pymatgen.core import Element

ElementTypeVar = TypeVar("ElementTypeVar", Element, str)

ElementType = Annotated[
    ElementTypeVar,
    BeforeValidator(lambda x: Element(x) if isinstance(x, str) else x),
    WrapSerializer(lambda x, nxt, info: str(x), return_type=str),
]
